from django.test import TestCase
from unittest.mock import patch
from datetime import datetime, timedelta
from dateutil.parser import parse

from .factories import ObservingRecordFactory, SiderealTargetFactory
from tom_observations.models import ObservationGroup, DynamicCadence
from tom_observations.cadences.resume_cadence_after_failure import ResumeCadenceAfterFailureStrategy
from tom_observations.cadences.retry_failed_observations import RetryFailedObservationsStrategy


mock_filters = {'1M0-SCICAM-SINISTRO': {
                    'type': 'IMAGE',
                    'class': '1m0',
                    'name': '1.0 meter Sinistro',
                    'optical_elements': {
                        'filters': [{'name': 'Bessell-I', 'code': 'I'}]}
                    }
                }

obs_params = {
        'facility': 'LCO',
        'observation_type': 'IMAGING',
        'name': 'With Perms',
        'ipp_value': 1.05,
        'start': '2020-01-01T00:00:00',
        'end': '2020-01-02T00:00:00',
        'exposure_count': 1,
        'exposure_time': 2.0,
        'max_airmass': 4.0,
        'observation_mode': 'NORMAL',
        'proposal': 'LCOSchedulerTest',
        'filter': 'I',
        'instrument_type': '1M0-SCICAM-SINISTRO'
    }


@patch('tom_observations.facilities.lco.LCOBaseForm._get_instruments', return_value=mock_filters)
@patch('tom_observations.facilities.lco.LCOBaseForm.proposal_choices',
       return_value=[('LCOSchedulerTest', 'LCOSchedulerTest')])
@patch('tom_observations.facilities.lco.LCOFacility.submit_observation', return_value=[198132])
@patch('tom_observations.facilities.lco.LCOFacility.validate_observation')
class TestReactiveCadencing(TestCase):
    def setUp(self):
        target = SiderealTargetFactory.create()
        obs_params['target_id'] = target.id
        obs_params['start'] = (datetime.now() - timedelta(hours=12)).strftime('%Y-%m-%dT%H:%M:%S')
        obs_params['end'] = (datetime.now() + timedelta(hours=12)).strftime('%Y-%m-%dT%H:%M:%S')
        observing_records = ObservingRecordFactory.create_batch(5,
                                                                target_id=target.id,
                                                                parameters=obs_params)
        self.group = ObservationGroup.objects.create()
        self.group.observation_records.add(*observing_records)
        self.group.save()
        self.dynamic_cadence = DynamicCadence.objects.create(
            cadence_strategy='Test Strategy', cadence_parameters={'cadence_frequency': 72}, active=True,
            observation_group=self.group)

    def test_retry_when_failed_cadence(self, patch1, patch2, patch3, patch4):
        num_records = self.group.observation_records.count()
        observing_record = self.group.observation_records.first()
        observing_record.status = 'CANCELED'
        observing_record.save()

        strategy = RetryFailedObservationsStrategy(self.dynamic_cadence)
        new_records = strategy.run()
        self.group.refresh_from_db()
        # Make sure the candence run created a new observation.
        self.assertEqual(num_records + 1, self.group.observation_records.count())
        # assert that the newly added observation record has a window of exactly 3 days
        # later than the canceled observation.
        self.assertEqual(
            parse(observing_record.parameters['start']),
            parse(new_records[0].parameters['start']) - timedelta(days=3)
        )

    @patch('tom_observations.facilities.lco.LCOFacility.get_observation_status', return_value={'state': 'CANCELED',
           'scheduled_start': None, 'scheduled_end': None})
    def test_resume_when_failed_cadence_failed_obs(self, mock_get_obs_status, mock_validate_obs, mock_submit_obs,
                                                   mock_proposal_choices, mock_get_insts):
        mock_validate_obs.return_value = []
        num_records = self.group.observation_records.count()

        strategy = ResumeCadenceAfterFailureStrategy(self.dynamic_cadence)
        new_records = strategy.run()
        self.group.refresh_from_db()
        self.assertEqual(num_records + 1, self.group.observation_records.count())
        self.assertEqual(
            datetime.now().replace(second=0, microsecond=0),
            parse(new_records[0].parameters['start']).replace(second=0, microsecond=0)
        )

    @patch('tom_observations.facilities.lco.LCOFacility.get_observation_status', return_value={'state': 'COMPLETED',
           'scheduled_start': None, 'scheduled_end': None})
    def test_resume_when_failed_cadence_successful_obs(self, mock_get_obs_status, mock_validate_obs, mock_submit_obs,
                                                       mock_proposal_choices, mock_get_insts):
        mock_validate_obs.return_value = []
        num_records = self.group.observation_records.count()
        obsr = self.group.observation_records.order_by('-created').first()

        strategy = ResumeCadenceAfterFailureStrategy(self.dynamic_cadence)
        new_records = strategy.run()
        self.group.refresh_from_db()
        self.assertEqual(num_records + 1, self.group.observation_records.count())
        self.assertAlmostEqual(
            parse(obsr.parameters['start']).replace(second=0, microsecond=0),
            parse(new_records[0].parameters['start']).replace(second=0, microsecond=0) - timedelta(days=3)
        )

    @patch('tom_observations.facilities.lco.LCOFacility.get_observation_status', return_value={'state': 'COMPLETED',
           'scheduled_start': None, 'scheduled_end': None})
    def test_resume_when_failed_cadence_invalid_date(self, mock_get_obs_status, mock_validate_obs, mock_submit_obs,
                                                     mock_proposal_choices, mock_get_insts):
        mock_validate_obs.return_value = []
        num_records = self.group.observation_records.count()
        obsr = self.group.observation_records.order_by('-created').first()
        obsr.parameters['start'] = (datetime.now() - timedelta(days=7)).strftime('%Y-%m-%dT%H:%M:%S')
        obsr.parameters['end'] = (datetime.now() - timedelta(days=6)).strftime('%Y-%m-%dT%H:%M:%S')
        obsr.save()

        strategy = ResumeCadenceAfterFailureStrategy(self.dynamic_cadence)
        new_records = strategy.run()
        self.group.refresh_from_db()
        self.assertEqual(num_records + 1, self.group.observation_records.count())
        self.assertAlmostEqual(
            datetime.now().replace(second=0, microsecond=0),
            parse(new_records[0].parameters['start']).replace(second=0, microsecond=0)
        )

    @patch('tom_observations.facilities.lco.LCOFacility.get_observation_status', return_value={'state': 'COMPLETED',
           'scheduled_start': None, 'scheduled_end': None})
    def test_resume_when_failed_cadence_obs_invalid(self, mock_get_obs_status, mock_validate_obs, mock_submit_obs,
                                                    mock_proposal_choices, mock_get_insts):
        mock_validate_obs.return_value = {'end': 'Window end time must be in the future'}

        strategy = ResumeCadenceAfterFailureStrategy(self.dynamic_cadence)
        with self.assertRaises(Exception):
            strategy.run()



# Generated by Django 2.2.1 on 2019-09-23 18:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0013_manual_20190916_multiple_names'),
    ]

    operations = [
        migrations.AddField(
            model_name='targetlist',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The time which this target was changed in the TOM database.', verbose_name='Last Modified'),
        ),
        migrations.AddField(
            model_name='targetname',
            name='modified',
            field=models.DateTimeField(auto_now=True, help_text='The time which this target was changed in the TOM database.', verbose_name='Last Modified'),
        ),
        migrations.AlterField(
            model_name='target',
            name='name',
            field=models.CharField(default='', help_text="The name of this target e.g. Barnard's star.", max_length=100, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='targetname',
            name='name',
            field=models.CharField(max_length=100, unique=True, verbose_name='Alias for target'),
        ),
    ]



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='tomtoolkit',
    description='The TOM Toolkit and base modules',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://tomtoolkit.github.io',
    author='TOM Toolkit Project',
    author_email='dcollom@lco.global',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics'
    ],
    keywords=['tomtoolkit', 'astronomy', 'astrophysics', 'cosmology', 'science', 'fits', 'observatory'],
    packages=find_packages(),
    use_scm_version=True,
    setup_requires=['setuptools_scm', 'wheel'],
    install_requires=[
        'astroquery>=0.4.2',
        'astroplan==0.8',
        'astropy==4.2.1',
        'beautifulsoup4~=4.9',
        'django~=3.1',  # TOM Toolkit requires db math functions
        'djangorestframework~=3.12',
        'django-bootstrap4~=3.0',
        'django-contrib-comments~=2.0',  # Earlier version are incompatible with Django >= 3.0
        'django-crispy-forms~=1.11',
        'django-extensions~=3.1',
        'django-filter~=2.4',
        'django-gravatar2~=1.4',
        'django-guardian~=2.3',
        'fits2image==0.4.4',
        'Markdown==3.3.4',  # django-rest-framework doc headers require this to support Markdown
        'numpy~=1.20',
        'pillow==8.3.2',
        'plotly~=5.0',
        'python-dateutil~=2.8',
        'requests~=2.25',
        'specutils==1.4.0',
    ],
    extras_require={
        'test': ['factory_boy==3.2.0'],
        'docs': [
            'recommonmark~=0.7',
            'sphinx~=4.0',
            'tom_antares',
            'tom_scimma'
        ]
    },
    include_package_data=True,
)



from tom_base.settings import *  # noqa

DATABASES = {
   'default': {
       'ENGINE': 'django.db.backends.postgresql',
       'NAME': 'tom',
       'USER': 'postgres',
       'PASSWORD': 'postgres',
       'HOST': '127.0.0.1',
       'PORT': 5432
   }
}



from django import forms
from django.urls import reverse
from crispy_forms.helper import FormHelper
from crispy_forms.layout import ButtonHolder, Column, Layout, Row, Submit

from tom_observations.facility import get_service_classes


def facility_choices():
    return [(k, k) for k in get_service_classes().keys()]


class AddExistingObservationForm(forms.Form):
    """
    This form is used for adding existing API-based observations to a Target object.
    """
    target_id = forms.IntegerField(required=True, widget=forms.HiddenInput())
    facility = forms.ChoiceField(required=True, choices=facility_choices, label=False)
    observation_id = forms.CharField(required=True, label=False,
                                     widget=forms.TextInput(attrs={'placeholder': 'Observation ID'}))
    confirm = forms.BooleanField(widget=forms.HiddenInput(), required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_action = reverse('tom_observations:add-existing')
        self.helper.layout = Layout(
            'target_id',
            'confirm',
            Row(
                Column(
                    'facility'
                ),
                Column(
                    'observation_id'
                ),
                Column(
                    ButtonHolder(
                        Submit('submit', 'Add Existing Observation')
                    )
                )
            )
        )


class UpdateObservationId(forms.Form):
    """
    This form is used for updating the observation ID on an ObservationRecord object.
    """
    obsr_id = forms.IntegerField(required=True, widget=forms.HiddenInput())
    observation_id = forms.CharField(required=True, label=False,
                                     widget=forms.TextInput(attrs={'placeholder': 'Observation ID'}))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_action = reverse('tom_observations:update', kwargs={'pk': self.initial.get('obsr_id')})
        self.helper.layout = Layout(
            'obsr_id',
            Row(
                Column(
                    'observation_id'
                ),
                Column(
                    ButtonHolder(
                        Submit('submit', 'Update Observation Id')
                    ),
                )
            )
        )



from django.contrib import admin
from .models import Target, TargetList, TargetExtra


class TargetExtraInline(admin.TabularInline):
    model = TargetExtra
    exclude = ('float_value', 'bool_value', 'time_value')
    extra = 0


class TargetAdmin(admin.ModelAdmin):
    model = Target
    inlines = [TargetExtraInline]


class TargetListAdmin(admin.ModelAdmin):
    model = TargetList


admin.site.register(Target, TargetAdmin)

admin.site.register(TargetList, TargetListAdmin)



# Generated by Django 2.2.9 on 2020-02-03 21:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_observations', '0005_auto_20191210_0007'),
    ]

    operations = [
        migrations.AddField(
            model_name='observationgroup',
            name='cadence_strategy',
            field=models.CharField(blank=True, default='', max_length=100),
        ),
    ]



from django.conf import settings
from django.contrib import messages

try:
    HINTS_ENABLED = settings.HINTS_ENABLED
    HINT_LEVEL = settings.HINT_LEVEL
except (AttributeError, KeyError):
    HINTS_ENABLED = False
    HINT_LEVEL = 20


def add_hint(request, hint_text, level=HINT_LEVEL):
    if HINTS_ENABLED:
        messages.add_message(request, level, hint_text)



# Generated by Django 2.2.2 on 2019-07-04 10:10

from django.db import migrations, models
import tom_dataproducts.models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_dataproducts', '0004_auto_20190626_0904'),
    ]

    operations = [
        migrations.AddField(
            model_name='dataproduct',
            name='thumbnail',
            field=models.FileField(default=None, null=True, upload_to=tom_dataproducts.models.data_product_path),
        ),
        migrations.AlterField(
            model_name='dataproduct',
            name='tag',
            field=models.CharField(blank=True, choices=[('photometry', 'Photometry'), ('fits_file', 'Fits File'), ('spectroscopy', 'Spectroscopy'), ('image_file', 'Image File')], default='', max_length=50),
        ),
        migrations.AlterField(
            model_name='reduceddatum',
            name='data_type',
            field=models.CharField(choices=[('spectroscopy', 'Spectroscopy'), ('photometry', 'Photometry'), ('image_file', 'Image File')], default='', max_length=100),
        ),
    ]



from django.contrib.auth.models import Group
from guardian.shortcuts import assign_perm, get_groups_with_perms, get_objects_for_user
from rest_framework import serializers

from tom_common.serializers import GroupSerializer
from tom_targets.models import Target, TargetExtra, TargetName
from tom_targets.validators import RequiredFieldsTogetherValidator


class TargetNameSerializer(serializers.ModelSerializer):
    id = serializers.IntegerField(required=False)

    class Meta:
        model = TargetName
        fields = ('id', 'name',)


class TargetExtraSerializer(serializers.ModelSerializer):
    id = serializers.IntegerField(required=False)

    class Meta:
        model = TargetExtra
        fields = ('id', 'key', 'value')


class TargetSerializer(serializers.ModelSerializer):
    """Target serializer responsbile for transforming models to/from
    json (or other representations). See
    https://www.django-rest-framework.org/api-guide/serializers/#modelserializer
    """
    targetextra_set = TargetExtraSerializer(many=True)
    aliases = TargetNameSerializer(many=True)
    groups = GroupSerializer(many=True, required=False)  # TODO: return groups in detail and list

    class Meta:
        model = Target
        fields = '__all__'
        # TODO: We should investigate if this validator logic can be reused in the forms to reduce code duplication.
        # TODO: Try to put validators in settings to allow user changes
        validators = [RequiredFieldsTogetherValidator('type', 'SIDEREAL', 'ra', 'dec'),
                      RequiredFieldsTogetherValidator('type', 'NON_SIDEREAL', 'epoch_of_elements', 'inclination',
                                                      'lng_asc_node', 'arg_of_perihelion', 'eccentricity'),
                      RequiredFieldsTogetherValidator('scheme', 'MPC_COMET', 'perihdist', 'epoch_of_perihelion'),
                      RequiredFieldsTogetherValidator('scheme', 'MPC_MINOR_PLANET', 'mean_anomaly', 'semimajor_axis'),
                      RequiredFieldsTogetherValidator('scheme', 'JPL_MAJOR_PLANET', 'mean_daily_motion', 'mean_anomaly',
                                                      'semimajor_axis')]

    def create(self, validated_data):
        """DRF requires explicitly handling writeable nested serializers,
        here we pop the alias/tag/group data and save it using their respective
        serializers
        """

        aliases = validated_data.pop('aliases', [])
        targetextras = validated_data.pop('targetextra_set', [])
        groups = validated_data.pop('groups', [])

        target = Target.objects.create(**validated_data)

        # Save groups for this target
        group_serializer = GroupSerializer(data=groups, many=True)
        if group_serializer.is_valid():
            for group in groups:
                group_instance = Group.objects.get(pk=group['id'])
                assign_perm('tom_targets.view_target', group_instance, target)
                assign_perm('tom_targets.change_target', group_instance, target)
                assign_perm('tom_targets.delete_target', group_instance, target)

        tns = TargetNameSerializer(data=aliases, many=True)
        if tns.is_valid():
            for alias in aliases:
                if alias['name'] == target.name:
                    target.delete()
                    alias_value = alias['name']
                    raise serializers.ValidationError(
                        f'Alias \'{alias_value}\' conflicts with Target name \'{target.name}\'.')
            tns.save(target=target)

        tes = TargetExtraSerializer(data=targetextras, many=True)
        if tes.is_valid():
            tes.save(target=target)

        return target

    def to_representation(self, instance):
        representation = super().to_representation(instance)
        groups = []
        for group in get_groups_with_perms(instance):
            groups.append(GroupSerializer(group).data)
        representation['groups'] = groups
        return representation

    def update(self, instance, validated_data):
        """
        For TargetExtra and TargetName objects, if the ID is present, it will update the corresponding row. If the ID is
        not present, it will attempt to create a new TargetExtra or TargetName associated with this Target.
        """
        aliases = validated_data.pop('aliases', [])
        targetextras = validated_data.pop('targetextra_set', [])
        groups = validated_data.pop('groups', [])

        # Save groups for this target
        group_serializer = GroupSerializer(data=groups, many=True)
        if group_serializer.is_valid():
            for group in groups:
                group_instance = Group.objects.get(pk=group['id'])
                assign_perm('tom_targets.view_target', group_instance, instance)
                assign_perm('tom_targets.change_target', group_instance, instance)
                assign_perm('tom_targets.delete_target', group_instance, instance)  # TODO: add tests

        for alias_data in aliases:
            alias = dict(alias_data)
            if alias['name'] == instance.name:  # Alias shouldn't conflict with target name
                alias_name = alias['name']
                raise serializers.ValidationError(
                    f'Alias \'{alias_name}\' conflicts with Target name \'{instance.name}\'.')
            if alias.get('id'):
                tn_instance = TargetName.objects.get(pk=alias['id'])
                if tn_instance.target != instance:  # Alias should correspond with target to be updated
                    raise serializers.ValidationError(f'''TargetName identified by id \'{tn_instance.id}\' is not an
                        alias of Target \'{instance.name}\'''')
                elif alias['name'] == tn_instance.name:
                    break  # Don't update if value doesn't change, because it will throw an error
                tns = TargetNameSerializer(tn_instance, data=alias_data)
            else:
                tns = TargetNameSerializer(data=alias_data)
            if tns.is_valid():
                tns.save(target=instance)

        for te_data in targetextras:
            te = dict(te_data)
            if te_data.get('id'):
                te_instance = TargetExtra.objects.get(pk=te['id'])
                tes = TargetExtraSerializer(te_instance, data=te_data)
            else:
                tes = TargetExtraSerializer(data=te_data)
            if tes.is_valid():
                tes.save(target=instance)

        fields_to_validate = ['name', 'type', 'ra', 'dec', 'epoch', 'parallax', 'pm_ra', 'pm_dec', 'galactic_lng',
                              'galactic_lat', 'distance', 'distance_err', 'scheme', 'epoch_of_elements',
                              'mean_anomaly', 'arg_of_perihelion', 'eccentricity', 'lng_asc_node', 'inclination',
                              'mean_daily_motion', 'semimajor_axis', 'epoch_of_perihelion', 'ephemeris_period',
                              'ephemeris_period_err', 'ephemeris_epoch', 'ephemeris_epoch_err', 'perihdist']
        for field in fields_to_validate:
            setattr(instance, field, validated_data.get(field, getattr(instance, field)))
        instance.save()

        return instance


class TargetFilteredPrimaryKeyRelatedField(serializers.PrimaryKeyRelatedField):
    # This PrimaryKeyRelatedField subclass is used to implement get_queryset based on the permissions of the user
    # submitting the request. The pattern was taken from this StackOverflow answer: https://stackoverflow.com/a/32683066

    def get_queryset(self):
        request = self.context.get('request', None)
        queryset = super().get_queryset()
        if not (request and queryset):
            return None
        return get_objects_for_user(request.user, 'tom_targets.change_target')



from django.contrib.auth.models import User
from django.db import models

from tom_targets.models import Target
from tom_observations.facility import get_service_class
from tom_common.hooks import run_hook


class ObservationRecord(models.Model):
    """
    Class representing an observation in a TOM.

    A ObservationRecord corresponds with any set of related exposures at a facility, and is associated with a single
    target.

    :param target: The ``Target`` with which this object is associated.
    :type target: Target

    :param facility: The facility at which this observation is taken. Should be the name specified in the corresponding
        TOM facility module, if one exists.
    :type facility: str

    :param parameters: The set of parameters used in the API request made to create the observation
    :type parameters: dict

    :param status: The current status of the observation. Should be a valid status in the corresponding TOM facility
        module, if one exists.
    :type status: str

    :param scheduled_start: The time at which the observation is scheduled to begin, according to the facility.
    :type scheduled_start: datetime

    :param scheduled_end: The time at which the observation is scheduled to end, according to the facility.
    :type scheduled_end: datetime

    :param created: The time at which this object was created.
    :type created: datetime

    :param modified: The time at which this object was last updated.
    :type modified: datetime
    """
    target = models.ForeignKey(Target, on_delete=models.CASCADE)
    user = models.ForeignKey(User, null=True, default=None, on_delete=models.DO_NOTHING)
    facility = models.CharField(max_length=50)
    parameters = models.JSONField()
    observation_id = models.CharField(max_length=255)
    status = models.CharField(max_length=200)
    scheduled_start = models.DateTimeField(null=True)
    scheduled_end = models.DateTimeField(null=True)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ('-created',)

    def save(self, *args, **kwargs):
        if self.id:
            presave_data = ObservationRecord.objects.get(pk=self.id)
            super().save(*args, **kwargs)
            if self.status != presave_data.status:
                run_hook('observation_change_state', self, presave_data.status)
        else:
            super().save(*args, **kwargs)
            run_hook('observation_change_state', self, None)

    @property
    def terminal(self):
        facility = get_service_class(self.facility)
        return self.status in facility().get_terminal_observing_states()

    @property
    def failed(self):
        facility = get_service_class(self.facility)
        return self.status in facility().get_failed_observing_states()

    @property
    def url(self):
        facility = get_service_class(self.facility)
        return facility().get_observation_url(self.observation_id)

    def update_status(self):
        facility = get_service_class(self.facility)
        facility().update_observation_status(self.id)

    def save_data(self):
        facility = get_service_class(self.facility)
        facility().save_data_products(self)

    def __str__(self):
        return '{0} @ {1}'.format(self.target, self.facility)


class ObservationGroup(models.Model):
    """
    Class representing a logical group of observations.

    :param name: The name of the grouping.
    :type name: str

    :param observation_records: Set of ``ObservationRecord`` objects associated with this ``ObservationGroup``

    :param created: The time at which this ``ObservationGroup`` was created.
    :type created: datetime

    :param modified: The time at which this ``ObservationGroup`` was modified.
    :type modified: datetime
    """
    name = models.CharField(max_length=50)
    observation_records = models.ManyToManyField(ObservationRecord)
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ('-created', 'name',)

    def __str__(self):
        return self.name


class DynamicCadence(models.Model):
    """
    Class representing a dynamic cadence--that is, a cadence that follows a pattern but modifies its behavior
    depending on the result of prior observations.

    :param observation_group: The ``ObservationGroup`` containing the observations that were created by this cadence.
    :type observation_group: ``ObservationGroup``

    :param cadence_strategy: The name of the cadence strategy this cadence is using.
    :type cadence_strategy: str

    :param cadence_parameters: The parameters for this cadence, e.g. cadence period
    :type cadence_parameters: JSON

    :param active: Whether or not this cadence should continue to submit observations
    :type active: boolean

    :param created: The time at which this ``DynamicCadence`` was created.
    :type created: datetime

    :param modified: The time at which this ``DynamicCadence`` was modified.
    :type modified: datetime
    """
    observation_group = models.ForeignKey(ObservationGroup, null=False, default=None, on_delete=models.CASCADE)
    cadence_strategy = models.CharField(max_length=100, blank=False, default=None,
                                        verbose_name='Cadence strategy used for this DynamicCadence')
    cadence_parameters = models.JSONField(blank=False, null=False, verbose_name='Cadence-specific parameters')
    active = models.BooleanField(verbose_name='Active',
                                 help_text='''Whether or not this DynamicCadence should
                                           continue to submit observations.''')
    created = models.DateTimeField(auto_now_add=True, help_text='The time which this DynamicCadence was created.')
    modified = models.DateTimeField(auto_now=True, help_text='The time which this DynamicCadence was modified.')

    def __str__(self):
        return f'{self.cadence_strategy} with parameters {self.cadence_parameters}'


class ObservationTemplate(models.Model):
    """
    Class representing an observation template.

    :param name: The name of the ``ObservationTemplate``
    :type name: str

    :param facility: The module-specified facility name for which the template is valid
    :type facility: str

    :param parameters: Observing parameters
    :type parameters: dict

    :param created: The time at which this ``ObservationTemplate`` was created.
    :type created: datetime

    :param modified: The time at which this ``ObservationTemplate`` was modified.
    :type modified: datetime
    """
    name = models.CharField(max_length=200)
    facility = models.CharField(max_length=50)
    parameters = models.JSONField()
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name



from datetime import timedelta

from django import forms
from django.utils import timezone
from astropy import units

from tom_observations.facility import BaseRoboticObservationFacility, GenericObservationForm
from tom_observations.facility import BaseManualObservationFacility
from tom_observations.models import ObservationRecord
from tom_observations.observation_template import GenericTemplateForm

# Site data matches built-in pyephem observer data for Los Angeles
SITES = {
    'Los Angeles': {
        'sitecode': 'lax',
        'latitude': 34.052222,
        'longitude': -117.756306,
        'elevation': 86.847092
    },
    'Siding Spring': {
        'sitecode': 'coj',
        'latitude': -31.272,
        'longitude': 149.07,
        'elevation': 1116
    },
}


class FakeFacilityForm(GenericObservationForm):
    test_input = forms.CharField(help_text='fake form input', required=True)


class FakeFacilityTemplateForm(GenericTemplateForm):
    pass


class FakeRoboticFacility(BaseRoboticObservationFacility):
    name = 'FakeRoboticFacility'
    observation_forms = {
        'OBSERVATION': FakeFacilityForm
    }

    def get_form(self, observation_type):
        return self.observation_forms[observation_type]

    def get_template_form(self, observation_type):
        return FakeFacilityTemplateForm

    def get_observing_sites(self):
        return SITES

    def get_observation_url(self, observation_id):
        return ''

    def data_products(self, observation_id, product_id=None):
        return [{'id': 'testdpid'}]

    def get_observation_status(self, observation_id):
        return {
            'state': 'COMPLETED',
            'scheduled_start': timezone.now() + timedelta(hours=1),
            'scheduled_end': timezone.now() + timedelta(hours=2)
        }

    def get_terminal_observing_states(self):
        return ['COMPLETED', 'FAILED', 'CANCELED', 'WINDOW_EXPIRED']

    def submit_observation(self, payload):
        return ['fakeid']

    def cancel_observation(self, observation_id):
        obsr = ObservationRecord.objects.get(observation_id=observation_id)
        obsr.status = 'CANCELED'
        obsr.save()
        return True

    def get_flux_constant(self):
        return units.erg / units.angstrom

    def get_wavelength_units(self):
        return units.angstrom

    def validate_observation(self, observation_payload):
        return True

    def get_facility_weather_urls(self):
        """
        `facility_weather_urls = {'code': 'XYZ', 'sites': [ site_dict, ... ]}`
        where
        `site_dict = {'code': 'XYZ', 'weather_url': 'http://path/to/weather'}`
        """
        # TODO: manually add a weather url for tlv
        facility_weather_urls = {
            'code': 'FakeRoboticFacility',
            'sites': [
                {
                    'code': site['sitecode'],
                    'weather_url': f'https://example.com/#/{site["sitecode"]}'
                }
                for site in SITES.values()]
            }

        return facility_weather_urls

    def get_facility_status(self):
        return {
            'code': 'LCO',
            'sites': [{'code': 'coj', 'telescopes': [{'code': 'coj.domb.1m0a', 'status': 'NOT_OK_TO_OPEN'}]}]
        }


class FakeManualFacility(BaseManualObservationFacility):
    name = 'FakeManualFacility'
    observation_forms = {
        'OBSERVATION': FakeFacilityForm
    }

    def get_form(self, observation_type):
        return FakeFacilityForm

    def get_template_form(self, observation_type):
        return FakeFacilityTemplateForm

    def get_observing_sites(self):
        return SITES

    def get_observation_url(self, observation_id):
        return ''

    def data_products(self, observation_id, product_id=None):
        return [{'id': 'testdpid'}]

    def get_observation_status(self, observation_id):
        return {
            'state': 'COMPLETED',
            'scheduled_start': timezone.now() + timedelta(hours=1),
            'scheduled_end': timezone.now() + timedelta(hours=2)
        }

    def get_terminal_observing_states(self):
        return ['COMPLETED', 'FAILED', 'CANCELED', 'WINDOW_EXPIRED']

    def submit_observation(self, payload):
        return ['fakeid']

    def get_flux_constant(self):
        return units.erg / units.angstrom

    def get_wavelength_units(self):
        return units.angstrom

    def validate_observation(self, observation_payload):
        return True

    # TOOD: this method does not belong to this Subclass of BaseObservationFacility
    # it's only here to satisfy tests.test_update_observations() which makes a (now)
    # invalid assumption that all facilities are robotic and have this method
    # The underlying problem is that when an ObservationRecord gets its facility
    # class, it assumes that it's a BaseRoboticFacility subclass.
    def update_all_observation_statuses(self, target):
        return []



from django import forms
from astropy.coordinates import Angle
from astropy import units as u
from django.forms import ValidationError, inlineformset_factory
from django.conf import settings
from django.contrib.auth.models import Group
from guardian.shortcuts import assign_perm, get_groups_with_perms, remove_perm

from .models import (
    Target, TargetExtra, TargetName, SIDEREAL_FIELDS, NON_SIDEREAL_FIELDS, REQUIRED_SIDEREAL_FIELDS,
    REQUIRED_NON_SIDEREAL_FIELDS, REQUIRED_NON_SIDEREAL_FIELDS_PER_SCHEME
)


def extra_field_to_form_field(field_type):
    if field_type == 'number':
        return forms.FloatField(required=False)
    elif field_type == 'boolean':
        return forms.BooleanField(required=False)
    elif field_type == 'datetime':
        return forms.DateTimeField(required=False)
    elif field_type == 'string':
        return forms.CharField(required=False, widget=forms.Textarea)
    else:
        raise ValueError(
            'Invalid field type {}. Field type must be one of: number, boolean, datetime string'.format(field_type)
        )


class CoordinateField(forms.CharField):
    def __init__(self, *args, **kwargs):
        c_type = kwargs.pop('c_type')
        self.c_type = c_type
        super().__init__(*args, **kwargs)

    def to_python(self, value):
        try:
            a = float(value)
            return a
        except ValueError:
            try:
                if self.c_type == 'ra':
                    a = Angle(value, unit=u.hourangle)
                else:
                    a = Angle(value, unit=u.degree)
                return a.to(u.degree).value
            except Exception:
                raise ValidationError('Invalid format. Please use sexigesimal or degrees')


class TargetForm(forms.ModelForm):
    groups = forms.ModelMultipleChoiceField(Group.objects.none(), required=False, widget=forms.CheckboxSelectMultiple)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.extra_fields = {}
        for extra_field in settings.EXTRA_FIELDS:
            # Add extra fields to the form
            field_name = extra_field['name']
            self.extra_fields[field_name] = extra_field_to_form_field(extra_field['type'])
            # Populate them with initial values if this is an update
            if kwargs['instance']:
                te = TargetExtra.objects.filter(target=kwargs['instance'], key=field_name)
                if te.exists():
                    self.extra_fields[field_name].initial = te.first().typed_value(extra_field['type'])

            self.fields.update(self.extra_fields)

    def save(self, commit=True):
        instance = super().save(commit=commit)
        if commit:
            for field in settings.EXTRA_FIELDS:
                if self.cleaned_data.get(field['name']) is not None:
                    TargetExtra.objects.update_or_create(
                            target=instance,
                            key=field['name'],
                            defaults={'value': self.cleaned_data[field['name']]}
                    )
            # Save groups for this target
            for group in self.cleaned_data['groups']:
                assign_perm('tom_targets.view_target', group, instance)
                assign_perm('tom_targets.change_target', group, instance)
                assign_perm('tom_targets.delete_target', group, instance)
            for group in get_groups_with_perms(instance):
                if group not in self.cleaned_data['groups']:
                    remove_perm('tom_targets.view_target', group, instance)
                    remove_perm('tom_targets.change_target', group, instance)
                    remove_perm('tom_targets.delete_target', group, instance)

        return instance

    class Meta:
        abstract = True
        model = Target
        fields = '__all__'
        widgets = {'type': forms.HiddenInput()}


class SiderealTargetCreateForm(TargetForm):
    ra = CoordinateField(required=True, label='Right Ascension', c_type='ra',
                         help_text='Right Ascension, in decimal degrees or sexagesimal hours. See '
                                   'https://docs.astropy.org/en/stable/api/astropy.coordinates.Angle.html for '
                                   'supported sexagesimal inputs.')
    dec = CoordinateField(required=True, label='Declination', c_type='dec',
                          help_text='Declination, in decimal or sexagesimal degrees. See '
                                    ' https://docs.astropy.org/en/stable/api/astropy.coordinates.Angle.html for '
                                    'supported sexagesimal inputs.')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in REQUIRED_SIDEREAL_FIELDS:
            self.fields[field].required = True

    class Meta(TargetForm.Meta):
        fields = SIDEREAL_FIELDS


class NonSiderealTargetCreateForm(TargetForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in REQUIRED_NON_SIDEREAL_FIELDS:
            self.fields[field].required = True

    def clean(self):
        """
        Look at the 'scheme' field and check the fields required for the
        specified field have been given
        """
        cleaned_data = super().clean()
        scheme = cleaned_data['scheme']  # scheme is a required field, so this should be safe
        required_fields = REQUIRED_NON_SIDEREAL_FIELDS_PER_SCHEME[scheme]

        for field in required_fields:
            if not cleaned_data.get(field):
                # Get verbose names of required fields
                field_names = [
                    "'" + Target._meta.get_field(f).verbose_name + "'"
                    for f in required_fields
                ]
                scheme_name = dict(Target.TARGET_SCHEMES)[scheme]
                raise ValidationError(
                    "Scheme '{}' requires fields {}".format(scheme_name, ', '.join(field_names))
                )

    class Meta(TargetForm.Meta):
        fields = NON_SIDEREAL_FIELDS


class TargetVisibilityForm(forms.Form):
    start_time = forms.DateTimeField(required=True, label='Start Time', widget=forms.TextInput(attrs={'type': 'date'}))
    end_time = forms.DateTimeField(required=True, label='End Time', widget=forms.TextInput(attrs={'type': 'date'}))
    airmass = forms.DecimalField(required=False, label='Maximum Airmass')

    def clean(self):
        cleaned_data = super().clean()
        start_time = cleaned_data.get('start_time')
        end_time = cleaned_data.get('end_time')
        target = self.data['target']
        if end_time < start_time:
            raise forms.ValidationError('Start time must be before end time')
        if target.type == 'NON_SIDEREAL':
            raise forms.ValidationError('Airmass plotting is only supported for sidereal targets')


TargetExtraFormset = inlineformset_factory(Target, TargetExtra, fields=('key', 'value'),
                                           widgets={'value': forms.TextInput()})
TargetNamesFormset = inlineformset_factory(Target, TargetName, fields=('name',), validate_min=False, can_delete=True,
                                           extra=3)



from django.contrib.auth.models import Group, User
from django.core.files.uploadedfile import SimpleUploadedFile
from django.urls import reverse
from guardian.shortcuts import assign_perm
from rest_framework import status
from rest_framework.test import APITestCase

from tom_dataproducts.models import DataProduct, ReducedDatum
from tom_observations.tests.factories import ObservingRecordFactory
from tom_targets.tests.factories import SiderealTargetFactory


class TestDataProductViewset(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username='testuser')
        self.client.force_login(self.user)
        self.st = SiderealTargetFactory.create()
        self.obsr = ObservingRecordFactory.create(target_id=self.st.id)
        self.dp_data = {
            'product_id': 'test_product_id',
            'target': self.st.id,
            'data_product_type': 'photometry'
        }

        assign_perm('tom_dataproducts.add_dataproduct', self.user)
        assign_perm('tom_targets.add_target', self.user, self.st)
        assign_perm('tom_targets.view_target', self.user, self.st)
        assign_perm('tom_targets.change_target', self.user, self.st)

    def test_data_product_upload_for_target(self):
        collaborator = User.objects.create(username='test collaborator')
        group = Group.objects.create(name='bourgeoisie')
        group.user_set.add(self.user)
        group.user_set.add(collaborator)

        with open('tom_dataproducts/tests/test_data/test_lightcurve.csv', 'rb') as lightcurve_file:
            self.dp_data['file'] = lightcurve_file
            response = self.client.post(reverse('api:dataproducts-list'), self.dp_data, format='multipart')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
            self.assertEqual(DataProduct.objects.count(), 1)
            self.assertEqual(ReducedDatum.objects.count(), 3)
            dp = DataProduct.objects.get(pk=response.data['id'])
            self.assertEqual(dp.target_id, self.st.id)

        # Test that group permissions are respected
        response = self.client.get(reverse('api:dataproducts-list'))
        self.assertContains(response, self.dp_data['product_id'], status_code=status.HTTP_200_OK)

    def test_data_product_upload_for_observation(self):
        self.dp_data['observation_record'] = self.obsr.id

        with open('tom_dataproducts/tests/test_data/test_lightcurve.csv', 'rb') as lightcurve_file:
            self.dp_data['file'] = lightcurve_file
            response = self.client.post(reverse('api:dataproducts-list'), self.dp_data, format='multipart')

            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
            self.assertEqual(DataProduct.objects.count(), 1)
            self.assertEqual(ReducedDatum.objects.count(), 3)
            dp = DataProduct.objects.get(pk=response.data['id'])
            self.assertEqual(dp.target_id, self.st.id)
            self.assertEqual(dp.observation_record_id, self.obsr.id)

    def test_data_product_upload_invalid_type(self):
        self.dp_data['data_product_type'] = 'invalid'

        with open('tom_dataproducts/tests/test_data/test_lightcurve.csv', 'rb') as lightcurve_file:
            self.dp_data['file'] = lightcurve_file
            response = self.client.post(reverse('api:dataproducts-list'), self.dp_data, format='multipart')

            self.assertContains(response, 'Not a valid data_product_type.', status_code=status.HTTP_400_BAD_REQUEST)

    def test_data_product_upload_failed_processing(self):
        self.dp_data['data_product_type'] = 'spectroscopy'

        with open('tom_dataproducts/tests/test_data/test_lightcurve.csv', 'rb') as lightcurve_file:
            self.dp_data['file'] = lightcurve_file
            response = self.client.post(reverse('api:dataproducts-list'), self.dp_data, format='multipart')

            self.assertContains(
                response,
                'There was an error in processing your DataProduct',
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
            )

    def test_data_product_delete(self):
        dp = DataProduct.objects.create(
            product_id='testproductid',
            target=self.st,
            data=SimpleUploadedFile('afile.fits', b'somedata')
        )
        assign_perm('tom_dataproducts.delete_dataproduct', self.user, dp)

        response = self.client.delete(reverse('api:dataproducts-detail', args=(dp.id,)))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_data_product_list(self):
        dp = DataProduct.objects.create(
            product_id='testproductid',
            target=self.st,
            data=SimpleUploadedFile('afile.fits', b'somedata')
        )

        response = self.client.get(reverse('api:dataproducts-list'))
        self.assertContains(response, dp.product_id, status_code=status.HTTP_200_OK)



# Generated by Django 2.1.7 on 2019-06-12 12:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0009_auto_20190605_1656'),
        ('tom_targets', '0008_merge_20190515_0914'),
    ]

    operations = [
    ]



import os

from django.core.files import File

from .models import DataProduct


def create_image_dataproduct(data_product):
    """
    Creates and saves a thumbnail image for a ``DataProduct``.

    :param data_product: ``DataProduct`` for which to create an image
    :type data_product: DataProduct

    :returns: True if creation was successful
    :rtype: boolean
    """
    tmpfile = data_product.create_thumbnail()
    if tmpfile:
        dp, _ = DataProduct.objects.get_or_create(
            product_id="{}_{}".format(data_product.product_id, "jpeg"),
            target=data_product.target,
            observation_record=data_product.observation_record,
            data_product_type='image_file',
        )
        outfile_name = os.path.basename(data_product.data.file.name)
        filename = outfile_name.split(".")[0] + ".jpg"
        with open(tmpfile.name, 'rb') as f:
            dp.data.save(filename, File(f), save=True)
            dp.save()
        tmpfile.close()
        return True

    return



# Generated by Django 3.0.7 on 2020-07-14 18:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0017_auto_20200130_2350'),
    ]

    operations = [
        migrations.AlterField(
            model_name='targetname',
            name='name',
            field=models.CharField(max_length=100, unique=True, verbose_name='Alias'),
        ),
    ]



# Generated by Django 2.1.5 on 2019-02-14 17:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0004_auto_20190123_2010'),
    ]

    operations = [
        migrations.AddField(
            model_name='targetextra',
            name='bool_value',
            field=models.BooleanField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='targetextra',
            name='float_value',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='targetextra',
            name='time_value',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



"""
Django settings for tom_base project.

Generated by 'django-admin startproject' using Django 2.0.6.

For more information on this file, see
https://docs.djangoproject.com/en/2.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.0/ref/settings/
"""
import logging.config
import os
import tempfile


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.getenv('SECRET_KEY', 'testkey')

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ['']


# Application definition

TOM_NAME = 'TOM Toolkit'

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',
    'django_extensions',
    'guardian',
    'tom_common',
    'django_comments',
    'bootstrap4',
    'crispy_forms',
    'rest_framework',
    'rest_framework.authtoken',
    'django_filters',
    'django_gravatar',
    'tom_targets',
    'tom_alerts',
    'tom_catalogs',
    'tom_observations',
    'tom_dataproducts',
]

SITE_ID = 1

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'tom_common.middleware.Raise403Middleware',
    'tom_common.middleware.ExternalServiceMiddleware',
    'tom_common.middleware.AuthStrategyMiddleware',
]

ROOT_URLCONF = 'tom_common.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

CRISPY_TEMPLATE_PACK = 'bootstrap4'

WSGI_APPLICATION = 'tom_base.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.0/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

# Password validation
# https://docs.djangoproject.com/en/2.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

LOGIN_URL = '/accounts/login/'
LOGIN_REDIRECT_URL = '/'
LOGOUT_REDIRECT_URL = '/'

AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',
    'guardian.backends.ObjectPermissionBackend',
)

# Internationalization
# https://docs.djangoproject.com/en/2.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = False

USE_TZ = True

DATETIME_FORMAT = 'Y-m-d H:i:s'
DATE_FORMAT = 'Y-m-d'


# Caching
# https://docs.djangoproject.com/en/dev/topics/cache/#filesystem-caching

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.filebased.FileBasedCache',
        'LOCATION': tempfile.gettempdir()
    }
}


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.0/howto/static-files/

STATIC_URL = '/static/'
MEDIA_ROOT = os.path.join(BASE_DIR, 'data')
MEDIA_URL = '/data/'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
        }
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'INFO'
        }
    }
}
logging.config.dictConfig(LOGGING)

TARGET_TYPE = 'SIDEREAL'
FACILITIES = {
    'LCO': {
        'portal_url': 'https://observe.lco.global',
        'api_key': os.getenv('LCO_API_KEY', ''),
    },
}

# Define the valid data product types for your TOM. Be careful when removing items, as previously valid types will no
# longer be valid, and may cause issues unless the offending records are modified.
DATA_PRODUCT_TYPES = {
    'photometry': ('photometry', 'Photometry'),
    'fits_file': ('fits_file', 'FITS File'),
    'spectroscopy': ('spectroscopy', 'Spectroscopy'),
    'image_file': ('image_file', 'Image File')
}

DATA_PROCESSORS = {
    'photometry': 'tom_dataproducts.processors.photometry_processor.PhotometryProcessor',
    'spectroscopy': 'tom_dataproducts.processors.spectroscopy_processor.SpectroscopyProcessor',
}

TOM_FACILITY_CLASSES = [
    'tom_observations.facilities.lco.LCOFacility',
    'tom_observations.facilities.gemini.GEMFacility',
    'tom_observations.facilities.soar.SOARFacility',
]

TOM_CADENCE_STRATEGIES = [
    'tom_observations.cadences.retry_failed_observations.RetryFailedObservationsStrategy',
    'tom_observations.cadences.resume_cadence_after_failure.ResumeCadenceAfterFailureStrategy'
]

# Define extra target fields here. Types can be any of "number", "string", "boolean" or "datetime"
# See https://tomtoolkit.github.io/docs/target_fields for documentation on this feature
# For example:
# EXTRA_FIELDS = [
#     {'name': 'redshift', 'type': 'number', 'default': 0},
#     {'name': 'discoverer', 'type': 'string'},
#     {'name': 'eligible', 'type': 'boolean', 'hidden': True},
#     {'name': 'dicovery_date', 'type': 'datetime'}
# ]
EXTRA_FIELDS = []

# Define custom DataProcessor class
# DATA_PROCESSOR_CLASS = 'mytom.custom_data_processor.CustomDataProcessor'

# Authentication strategy can either be LOCKED (required login for all views)
# or READ_ONLY (read only access to views)
AUTH_STRATEGY = 'READ_ONLY'

# Row-level data permissions restrict users from viewing certain objects unless they are a member of the group to which
# the object belongs. Setting this value to True will allow all `ObservationRecord`, `DataProduct`, and `ReducedDatum`
# objects to be seen by everyone. Setting it to False will allow users to specify which groups can access
# `ObservationRecord`, `DataProduct`, and `ReducedDatum` objects.
TARGET_PERMISSIONS_ONLY = True

# URLs that should be allowed access even with AUTH_STRATEGY = LOCKED
# for example: OPEN_URLS = ['/', '/about']
OPEN_URLS = []

HOOKS = {
    'target_post_save': 'tom_common.hooks.target_post_save',
    'observation_change_state': 'tom_common.hooks.observation_change_state',
    'data_product_post_upload': 'tom_dataproducts.hooks.data_product_post_upload'
}

AUTO_THUMBNAILS = False

THUMBNAIL_MAX_SIZE = (0, 0)

THUMBNAIL_DEFAULT_SIZE = (200, 200)

HINTS_ENABLED = False
HINT_LEVEL = 20

REST_FRAMEWORK = {
    'DEFAULT_PERMISSION_CLASSES': [
    ],
    'TEST_REQUEST_DEFAULT_FORMAT': 'json',
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.LimitOffsetPagination',
    'PAGE_SIZE': 100
}

try:
    from local_settings import *  # noqa
except ImportError:
    pass



from datetime import datetime, timedelta
import json
from requests import Response
from unittest.mock import patch

from django.test import TestCase

from tom_common.exceptions import ImproperCredentialsException
from tom_observations.facilities.lco import make_request, LCOFacility
from tom_observations.facilities.lco import LCOBaseForm, LCOBaseObservationForm, LCOImagingObservationForm
from tom_observations.facilities.lco import LCOPhotometricSequenceForm, LCOSpectroscopicSequenceForm
from tom_observations.facilities.lco import LCOSpectroscopyObservationForm, LCOMuscatImagingObservationForm
from tom_observations.tests.factories import SiderealTargetFactory, NonSiderealTargetFactory


instrument_response = {
    '2M0-FLOYDS-SCICAM': {
        'type': 'SPECTRA', 'class': '2m0', 'name': '2.0 meter FLOYDS', 'optical_elements': {
            'slits': [
              {'name': '6.0 arcsec slit', 'code': 'slit_6.0as', 'schedulable': True, 'default': False},
              {'name': '1.6 arcsec slit', 'code': 'slit_1.6as', 'schedulable': True, 'default': False},
              {'name': '2.0 arcsec slit', 'code': 'slit_2.0as', 'schedulable': True, 'default': False},
              {'name': '1.2 arcsec slit', 'code': 'slit_1.2as', 'schedulable': True, 'default': False}
            ]
        }
    },
    '0M4-SCICAM-SBIG': {
        'type': 'IMAGE', 'class': '0m4', 'name': '0.4 meter SBIG', 'optical_elements': {
            'filters': [
                {"name": "Bessell-U", "code": "U", "schedulable": True, "default": False},
                {"name": "Bessell-B", "code": "B", "schedulable": True, "default": False},
                {"name": "Bessell-V", "code": "V", "schedulable": True, "default": False},
                {"name": "Bessell-R", "code": "R", "schedulable": True, "default": False},
                {"name": "Bessell-I", "code": "I", "schedulable": True, "default": False},
                {"name": "SDSS-up", "code": "up", "schedulable": True, "default": False},
                {"name": "SDSS-gp", "code": "gp", "schedulable": True, "default": False},
                {"name": "SDSS-rp", "code": "rp", "schedulable": True, "default": False},
                {"name": "SDSS-ip", "code": "ip", "schedulable": True, "default": False},
                {"name": "PanSTARRS-Z", "code": "zs", "schedulable": True, "default": False},
                {"name": "PanSTARRS-w", "code": "w", "schedulable": True, "default": False},
                {'name': 'Opaque', 'code': 'opaque', 'schedulable': False, 'default': False},
                {'name': '100um Pinhole', 'code': '100um-Pinhole', 'schedulable': False, 'default': False},
            ]
        },
        'modes': {
            'guiding': {
                'type': 'guiding',
                'modes': [
                    {'name': 'On', 'overhead': 0.0, 'code': 'ON'},
                    {'name': 'Off', 'overhead': 0.0, 'code': 'OFF'},
                ],
                'default': 'ON'
            }
        }
    },
    'SOAR_GHTS_REDCAM': {
        'type': 'SPECTRA', 'class': '4m0', 'name': 'Goodman Spectrograph RedCam', 'optical_elements': {
            'gratings': [
                {'name': '400 line grating', 'code': 'SYZY_400', 'schedulable': True, 'default': True},
            ],
            'slits': [
                {'name': '1.0 arcsec slit', 'code': 'slit_1.0as', 'schedulable': True, 'default': True}
            ]
        },
    },
    '2M0-SCICAM-MUSCAT': {
        'type': 'IMAGE', 'class': '2m0', 'name': '2.0 meter Muscat',
        'optical_elements': {
            'diffuser_g_positions': [
                {'name': 'In Beam', 'code': 'in', 'schedulable': True, 'default': True},
                {'name': 'Out of Beam', 'code': 'out', 'schedulable': True, 'default': True}
            ],
            'diffuser_r_positions': [
                {'name': 'In Beam', 'code': 'in', 'schedulable': True, 'default': False},
                {'name': 'Out of Beam', 'code': 'out', 'schedulable': True, 'default': True}
            ],
            'diffuser_i_positions': [
                {'name': 'In Beam', 'code': 'in', 'schedulable': True, 'default': False},
                {'name': 'Out of Beam', 'code': 'out', 'schedulable': True, 'default': True}
            ],
            'diffuser_z_positions': [
                {'name': 'In Beam', 'code': 'in', 'schedulable': True, 'default': False},
                {'name': 'Out of Beam', 'code': 'out', 'schedulable': True, 'default': True}
            ]
        },
        'modes': {
            'guiding': {
                'type': 'guiding',
                'modes': [
                    {'name': 'On', 'overhead': 0.0, 'code': 'ON'},
                    {'name': 'Muscat G Guiding', 'overhead': 0.0, 'code': 'MUSCAT_G'},
                ],
                'default': 'ON'
            },
            'exposure': {
                'type': 'exposure',
                'modes': [
                    {'name': 'Muscat Synchronous Exposure Mode', 'overhead': 0.0, 'code': 'SYNCHRONOUS'},
                    {'name': 'Muscat Asynchronous Exposure Mode', 'overhead': 0.0, 'code': 'ASYNCHRONOUS'}
                ],
                'default': 'SYNCHRONOUS'
            }
        }
    }
}


def generate_lco_instrument_choices():
    return {k: v for k, v in instrument_response.items() if 'SOAR' not in k}


def generate_lco_proposal_choices():
    return [('sampleproposal', 'Sample Proposal')]


class TestMakeRequest(TestCase):

    @patch('tom_observations.facilities.lco.requests.request')
    def test_make_request(self, mock_request):
        mock_response = Response()
        mock_response._content = str.encode(json.dumps({'test': 'test'}))
        mock_response.status_code = 200
        mock_request.return_value = mock_response

        self.assertDictEqual({'test': 'test'}, make_request('GET', 'google.com', headers={'test': 'test'}).json())

        mock_response.status_code = 403
        mock_request.return_value = mock_response
        with self.assertRaises(ImproperCredentialsException):
            make_request('GET', 'google.com', headers={'test': 'test'})


class TestLCOBaseForm(TestCase):

    @patch('tom_observations.facilities.lco.make_request')
    @patch('tom_observations.facilities.lco.cache')
    def test_get_instruments(self, mock_cache, mock_make_request):
        mock_response = Response()
        mock_response._content = str.encode(json.dumps(instrument_response))
        mock_response.status_code = 200
        mock_make_request.return_value = mock_response

        # Test that cached value is returned
        with self.subTest():
            test_instruments = {'test instrument': {'type': 'IMAGE'}}
            mock_cache.get.return_value = test_instruments

            instruments = LCOBaseForm._get_instruments()
            self.assertDictContainsSubset({'test instrument': {'type': 'IMAGE'}}, instruments)
            self.assertNotIn('0M4-SCICAM-SBIG', instruments)

        # Test that empty cache results in mock_instruments, and cache.set is called
        with self.subTest():
            mock_cache.get.return_value = None

            instruments = LCOBaseForm._get_instruments()
            self.assertIn('0M4-SCICAM-SBIG', instruments)
            self.assertDictContainsSubset({'type': 'IMAGE'}, instruments['0M4-SCICAM-SBIG'])
            self.assertNotIn('SOAR_GHTS_REDCAM', instruments)
            mock_cache.set.assert_called()

    @patch('tom_observations.facilities.lco.LCOBaseForm._get_instruments')
    def test_instrument_choices(self, mock_get_instruments):
        mock_get_instruments.return_value = generate_lco_instrument_choices()

        inst_choices = LCOBaseForm.instrument_choices()
        self.assertIn(('2M0-FLOYDS-SCICAM', '2.0 meter FLOYDS'), inst_choices)
        self.assertIn(('0M4-SCICAM-SBIG', '0.4 meter SBIG'), inst_choices)
        self.assertIn(('2M0-SCICAM-MUSCAT', '2.0 meter Muscat'), inst_choices)
        self.assertEqual(len(inst_choices), 3)

    @patch('tom_observations.facilities.lco.LCOBaseForm._get_instruments')
    def test_filter_choices(self, mock_get_instruments):
        mock_get_instruments.return_value = generate_lco_instrument_choices()

        filter_choices = LCOBaseForm.filter_choices()
        for expected in [('opaque', 'Opaque'), ('100um-Pinhole', '100um Pinhole'), ('slit_6.0as', '6.0 arcsec slit')]:
            self.assertIn(expected, filter_choices)
        self.assertEqual(len(filter_choices), 17)

    @patch('tom_observations.facilities.lco.make_request')
    def test_proposal_choices(self, mock_make_request):
        mock_response = Response()
        mock_response._content = str.encode(json.dumps({'proposals': [
            {'id': 'ActiveProposal', 'title': 'Active', 'current': True},
            {'id': 'InactiveProposal', 'title': 'Inactive', 'current': False}]
        }))
        mock_response.status_code = 200
        mock_make_request.return_value = mock_response

        proposal_choices = LCOBaseForm.proposal_choices()
        self.assertIn(('ActiveProposal', 'Active (ActiveProposal)'), proposal_choices)
        self.assertNotIn(('InactiveProposal', 'Inactive (InactiveProposal)'), proposal_choices)


@patch('tom_observations.facilities.lco.LCOBaseObservationForm.proposal_choices')
@patch('tom_observations.facilities.lco.LCOBaseObservationForm.filter_choices')
@patch('tom_observations.facilities.lco.LCOBaseObservationForm.instrument_choices')
@patch('tom_observations.facilities.lco.LCOBaseObservationForm.validate_at_facility')
class TestLCOBaseObservationForm(TestCase):

    def setUp(self):
        self.st = SiderealTargetFactory.create()
        self.nst = NonSiderealTargetFactory.create(scheme='MPC_MINOR_PLANET')
        self.valid_form_data = {
            'name': 'test', 'facility': 'LCO', 'target_id': self.st.id, 'ipp_value': 0.5, 'start': '2020-11-03',
            'end': '2020-11-04', 'exposure_count': 1, 'exposure_time': 30, 'max_airmass': 3,
            'min_lunar_distance': 20, 'observation_mode': 'NORMAL',
            'proposal': 'sampleproposal', 'filter': 'opaque', 'instrument_type': '0M4-SCICAM-SBIG'
        }
        self.instrument_choices = [(k, v['name']) for k, v in instrument_response.items() if 'SOAR' not in k]
        self.filter_choices = set([
            (f['code'], f['name']) for ins in instrument_response.values() for f in
            ins['optical_elements'].get('filters', []) + ins['optical_elements'].get('slits', [])
        ])
        self.proposal_choices = generate_lco_proposal_choices()

    def test_validate_at_facility(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        pass

    def test_clean_and_validate(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        """Test clean_start, clean_end, and is_valid()"""
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        # Test that a valid form returns True, and that start and end are cleaned properly
        form = LCOBaseObservationForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual('2020-11-03T00:00:00', form.cleaned_data['start'])
        self.assertEqual('2020-11-04T00:00:00', form.cleaned_data['end'])

        # Test that an invalid form returns False
        self.valid_form_data.pop('target_id')
        form = LCOBaseObservationForm(self.valid_form_data)
        self.assertFalse(form.is_valid())

        # TODO: Add test for when validate_at_facility returns errors

    def test_flatten_error_dict(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        test_error = {
            'requests': [{'configurations': [
                {'non_field_errors': ['Test configuration error']},
                {'max_airmass': ['Invalid airmass']}
            ]}],
            'test_string': 'string error',
            'ipp_value': 'Invalid ipp',
            'test_dict': {'test_key': 'dict_error'}
        }
        form = LCOBaseObservationForm(self.valid_form_data)
        flattened_errors = form._flatten_error_dict(test_error)
        self.assertIn(['non_field_errors: Test configuration error'], flattened_errors[0])
        self.assertIn('test_string: string error', flattened_errors)
        self.assertIn('Invalid ipp', form.errors['ipp_value'])
        self.assertIn('Invalid airmass', form.errors['max_airmass'])
        self.assertIn(['test_key: dict_error'], flattened_errors)

    def test_instrument_to_type(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        """Test instrument_to_type method."""
        self.assertEqual('SPECTRUM', LCOBaseObservationForm.instrument_to_type('2M0-FLOYDS-SCICAM'))
        self.assertEqual('NRES_SPECTRUM', LCOBaseObservationForm.instrument_to_type('1M0-NRES-SCICAM'))
        self.assertEqual('EXPOSE', LCOBaseObservationForm.instrument_to_type('0M4-SCICAM-SBIG'))

    def test_build_target_fields(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        """Test _build_target_fields method."""
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        # Test correct population of target fields for a sidereal target
        with self.subTest():
            form = LCOBaseObservationForm(self.valid_form_data)
            self.assertTrue(form.is_valid())
            self.assertDictEqual({
                'name': self.st.name, 'type': 'ICRS', 'ra': self.st.ra, 'dec': self.st.dec,
                'proper_motion_ra': self.st.pm_ra, 'proper_motion_dec': self.st.pm_dec, 'epoch': self.st.epoch
            }, form._build_target_fields())

        # Test correct population of target fields for a non-sidereal target
        with self.subTest():
            self.valid_form_data['target_id'] = self.nst.id
            form = LCOBaseObservationForm(self.valid_form_data)
            self.assertTrue(form.is_valid())
            self.assertDictContainsSubset({
                'name': self.nst.name, 'type': 'ORBITAL_ELEMENTS', 'epochofel': self.nst.epoch_of_elements,
                'orbinc': self.nst.inclination, 'longascnode': self.nst.lng_asc_node,
                'argofperih': self.nst.arg_of_perihelion, 'meananom': self.nst.mean_anomaly,
                'meandist': self.nst.semimajor_axis
            }, form._build_target_fields())

    def test_build_instrument_config(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        """Test _build_instrument_config method."""
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        form = LCOBaseObservationForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(
            [{'exposure_count': self.valid_form_data['exposure_count'],
              'exposure_time': self.valid_form_data['exposure_time'],
              'optical_elements': {'filter': self.valid_form_data['filter']}
              }],
            form._build_instrument_config()
        )

    def test_build_acquisition_config(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        """Test _build_acquisition_config method."""
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        form = LCOBaseObservationForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        self.assertDictEqual({}, form._build_acquisition_config())

    def test_build_guiding_config(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        """Test _build_guiding_config method."""
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        form = LCOBaseObservationForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        self.assertDictEqual({}, form._build_guiding_config())

    # This should but does not mock instrument_to_type, _build_target_fields, _build_instrument_config,
    # _build_acquisition_config, and _build_guiding_config
    def test_build_configuration(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        """Test _build_configuration method."""
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        form = LCOBaseObservationForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        configuration = form._build_configuration()
        self.assertDictContainsSubset(
            {'type': 'EXPOSE', 'instrument_type': '0M4-SCICAM-SBIG',
             'constraints': {'max_airmass': 3, 'min_lunar_distance': 20}},
            configuration)
        for key in ['target', 'instrument_configs', 'acquisition_config', 'guiding_config']:
            self.assertIn(key, configuration)

    @patch('tom_observations.facilities.lco.LCOBaseForm._get_instruments')
    def test_build_location(self, mock_get_instruments, mock_validate, mock_insts, mock_filters, mock_proposals):
        """Test _build_location method."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        form = LCOBaseObservationForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        self.assertDictEqual({'telescope_class': '0m4'}, form._build_location())

    def test_expand_cadence_request(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        pass

    @patch('tom_observations.facilities.lco.LCOBaseObservationForm._build_location')
    @patch('tom_observations.facilities.lco.LCOBaseObservationForm._build_configuration')
    @patch('tom_observations.facilities.lco.make_request')
    def test_observation_payload(self, mock_make_request, mock_build_configuration, mock_build_location, mock_validate,
                                 mock_insts, mock_filters, mock_proposals):
        """Test observation_payload method."""
        mock_build_configuration.return_value = {}
        mock_build_location.return_value = {}
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        # Test a non-static cadenced form
        with self.subTest():
            form = LCOBaseObservationForm(self.valid_form_data)
            self.assertTrue(form.is_valid())
            obs_payload = form.observation_payload()
            self.assertDictContainsSubset(
                {'name': 'test', 'proposal': 'sampleproposal', 'ipp_value': 0.5, 'operator': 'SINGLE',
                 'observation_type': 'NORMAL'}, obs_payload
            )
            self.assertNotIn('cadence', obs_payload['requests'][0])

        # Test a static cadence form
        with self.subTest():
            mock_response = Response()
            mock_response._content = str.encode(json.dumps({'test': 'test_static_cadence'}))
            mock_response.status_code = 200
            mock_make_request.return_value = mock_response

            self.valid_form_data['period'] = 60
            self.valid_form_data['jitter'] = 15
            form = LCOBaseObservationForm(self.valid_form_data)
            self.assertTrue(form.is_valid())
            self.assertDictEqual({'test': 'test_static_cadence'}, form.observation_payload())


@patch('tom_observations.facilities.lco.LCOImagingObservationForm._get_instruments')
class TestLCOImagingObservationForm(TestCase):
    def test_instrument_choices(self, mock_get_instruments):
        """Test LCOImagingObservationForm._instrument_choices."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()

        inst_choices = LCOImagingObservationForm.instrument_choices()
        self.assertIn(('0M4-SCICAM-SBIG', '0.4 meter SBIG'), inst_choices)
        self.assertNotIn(('2M0-FLOYDS-SCICAM', '2.0 meter FLOYDS'), inst_choices)
        self.assertNotIn(('2M0-SCICAM-MUSCAT', '2.0 meter Muscat'), inst_choices)
        self.assertEqual(len(inst_choices), 1)

    def test_filter_choices(self, mock_get_instruments):
        """Test LCOImagingObservationForm._filter_choices."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()

        filter_choices = LCOImagingObservationForm.filter_choices()
        for expected in [('opaque', 'Opaque'), ('100um-Pinhole', '100um Pinhole')]:
            self.assertIn(expected, filter_choices)
        for not_expected in [('slit_6.0as', '6.0 arcsec slit'), ('slit_1.6as', '1.6 arcsec slit'),
                             ('slit_2.0as', '2.0 arcsec slit'), ('slit_1.2as', '1.2 arcsec slit')]:
            self.assertNotIn(not_expected, filter_choices)
        self.assertEqual(len(filter_choices), 13)


@patch('tom_observations.facilities.lco.LCOMuscatImagingObservationForm.validate_at_facility')
@patch('tom_observations.facilities.lco.LCOMuscatImagingObservationForm.filter_choices')
@patch('tom_observations.facilities.lco.LCOMuscatImagingObservationForm.proposal_choices')
@patch('tom_observations.facilities.lco.LCOMuscatImagingObservationForm._get_instruments')
class TestLCOMuscatImagingObservationForm(TestCase):

    def setUp(self):
        self.st = SiderealTargetFactory.create()
        self.valid_form_data = {
            'name': 'test', 'facility': 'LCO', 'target_id': self.st.id, 'ipp_value': 0.5, 'start': '2020-11-03',
            'end': '2020-11-04', 'exposure_count': 1, 'exposure_time_g': 30, 'exposure_time_r': 40,
            'exposure_time_i': 50, 'exposure_time_z': 60, 'diffuser_g_position': 'out', 'diffuser_r_position': 'out',
            'diffuser_i_position': 'in', 'diffuser_z_position': 'in', 'observation_mode': 'NORMAL',
            'guider_mode': 'MUSCAT_G', 'exposure_mode': 'SYNCHRONOUS', 'proposal': 'sampleproposal',
            'instrument_type': '2M0-SCICAM-MUSCAT', 'max_airmass': 3.0
        }

    def test_instrument_choices(self, mock_get_instruments, mock_get_proposals, mock_get_filters, mock_validate):
        """Test LCOMuscatImagingObservationForm._instrument_choices."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()
        inst_choices = LCOMuscatImagingObservationForm.instrument_choices()
        self.assertIn(('2M0-SCICAM-MUSCAT', '2.0 meter Muscat'), inst_choices)
        self.assertEqual(len(inst_choices), 1)

    def test_diffuser_choices(self, mock_get_instruments, mock_get_proposals, mock_get_filters, mock_validate):
        """Test LCOMuscatImagingObservationForm.diffuser_position_choices."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()
        for channel in ['g', 'r', 'i', 'z']:
            with self.subTest(channel=channel):
                diffuser_choices = LCOMuscatImagingObservationForm.diffuser_position_choices(channel=channel)
                self.assertIn(('in', 'In Beam'), diffuser_choices)
                self.assertIn(('out', 'Out of Beam'), diffuser_choices)
                self.assertTrue(len(diffuser_choices), 2)

    def test_exposure_mode_choices(self, mock_get_instruments, mock_get_proposals, mock_get_filters, mock_validate):
        """Test LCOMuscatImagingObservationForm.mode_choices for exposure modes."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()
        exposure_mode_choices = LCOMuscatImagingObservationForm.mode_choices('exposure')
        self.assertIn(('SYNCHRONOUS', 'Muscat Synchronous Exposure Mode'), exposure_mode_choices)
        self.assertIn(('ASYNCHRONOUS', 'Muscat Asynchronous Exposure Mode'), exposure_mode_choices)
        self.assertEqual(len(exposure_mode_choices), 2)

    def test_guide_mode_choices(self, mock_get_instruments, mock_get_proposals, mock_get_filters, mock_validate):
        """Test LCOMuscatImagingObservationForm.mode_choices for guiding modes."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()
        guide_mode_choices = LCOMuscatImagingObservationForm.mode_choices('guiding')
        self.assertIn(('MUSCAT_G', 'Muscat G Guiding'), guide_mode_choices)
        self.assertIn(('ON', 'On'), guide_mode_choices)
        self.assertEqual(len(guide_mode_choices), 2)

    def test_build_instrument_config(self, mock_get_instruments, mock_get_proposals, mock_get_filters, mock_validate):
        """Test LCOMuscatImagingObservationForm._build_instrument_config."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()
        mock_get_proposals.return_value = generate_lco_proposal_choices()
        form = LCOMuscatImagingObservationForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        instrument_config = form._build_instrument_config()
        expected_exposure_time = max(
            self.valid_form_data['exposure_time_g'],
            self.valid_form_data['exposure_time_r'],
            self.valid_form_data['exposure_time_i'],
            self.valid_form_data['exposure_time_z']
        )
        self.assertEqual(expected_exposure_time, instrument_config[0]['exposure_time'])
        self.assertDictEqual({
            'exposure_mode': self.valid_form_data['exposure_mode'],
            'exposure_time_g': self.valid_form_data['exposure_time_g'],
            'exposure_time_r': self.valid_form_data['exposure_time_r'],
            'exposure_time_i': self.valid_form_data['exposure_time_i'],
            'exposure_time_z': self.valid_form_data['exposure_time_z']
        }, instrument_config[0]['extra_params'])
        self.assertDictEqual({
            'diffuser_g_position': 'out',
            'diffuser_r_position': 'out',
            'diffuser_i_position': 'in',
            'diffuser_z_position': 'in'
        }, instrument_config[0]['optical_elements'])

    def test_build_guiding_config(self, mock_get_instruments, mock_get_proposals, mock_get_filters, mock_validate):
        """Test LCOMuscatImagingObservationForm._build_guiding_config."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()
        mock_get_proposals.return_value = generate_lco_proposal_choices()
        form = LCOMuscatImagingObservationForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        guiding_config = form._build_guiding_config()
        self.assertDictEqual({'mode': 'MUSCAT_G', 'optional': True},  guiding_config)


class TestLCOSpectroscopyObservationForm(TestCase):
    @patch('tom_observations.facilities.lco.LCOSpectroscopyObservationForm._get_instruments')
    def test_instrument_choices(self, mock_get_instruments):
        """Test LCOSpectroscopyObservationForm._instrument_choices."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()

        inst_choices = LCOSpectroscopyObservationForm.instrument_choices()
        self.assertIn(('2M0-FLOYDS-SCICAM', '2.0 meter FLOYDS'), inst_choices)
        self.assertNotIn(('0M4-SCICAM-SBIG', '0.4 meter SBIG'), inst_choices)
        self.assertEqual(len(inst_choices), 1)

    @patch('tom_observations.facilities.lco.LCOSpectroscopyObservationForm._get_instruments')
    def test_filter_choices(self, mock_get_instruments):
        """Test LCOSpectroscopyObservationForm._filter_choices."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()

        filter_choices = LCOSpectroscopyObservationForm.filter_choices()
        for expected in [('slit_6.0as', '6.0 arcsec slit'), ('slit_1.6as', '1.6 arcsec slit'),
                         ('slit_2.0as', '2.0 arcsec slit'), ('slit_1.2as', '1.2 arcsec slit'), ('None', 'None')]:
            self.assertIn(expected, filter_choices)
        for not_expected in [('opaque', 'Opaque'), ('100um-Pinhole', '100um Pinhole')]:
            self.assertNotIn(not_expected, filter_choices)
        self.assertEqual(len(filter_choices), 5)

    @patch('tom_observations.facilities.lco.LCOSpectroscopyObservationForm.proposal_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopyObservationForm.filter_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopyObservationForm.instrument_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopyObservationForm.validate_at_facility')
    def test_build_instrument_config(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        mock_validate.return_value = []
        mock_insts.return_value = [(k, v['name']) for k, v in instrument_response.items() if 'SPECTRA' in v['type']]
        mock_filters.return_value = set([
            (f['code'], f['name']) for ins in instrument_response.values() for f in
            ins['optical_elements'].get('slits', [])
            ] + [('None', 'None')])
        mock_proposals.return_value = generate_lco_proposal_choices()

        st = SiderealTargetFactory.create()
        valid_form_data = {
            'name': 'test', 'facility': 'LCO', 'target_id': st.id, 'ipp_value': 0.5, 'start': '2020-11-03',
            'end': '2020-11-04', 'exposure_count': 1, 'exposure_time': 30.0, 'max_airmass': 3.0,
            'min_lunar_distance': 20, 'observation_mode': 'NORMAL', 'proposal': 'sampleproposal',
            'filter': 'slit_2.0as', 'instrument_type': '2M0-FLOYDS-SCICAM', 'rotator_angle': 1.0
        }

        # Test that optical_elements['slit'] is populated when filter is included
        with self.subTest():
            form = LCOSpectroscopyObservationForm(valid_form_data)
            self.assertTrue(form.is_valid())
            self.assertEqual(
                [{'exposure_count': valid_form_data['exposure_count'],
                  'exposure_time': valid_form_data['exposure_time'],
                  'optical_elements': {'slit': valid_form_data['filter']},
                  'rotator_mode': 'VFLOAT',
                  'extra_params': {'rotator_angle': valid_form_data['rotator_angle']}
                  }], form._build_instrument_config()
            )

        # Test that optical elements is removed when filter is excluded
        with self.subTest():
            valid_form_data['filter'] = 'None'
            form = LCOSpectroscopyObservationForm(valid_form_data)
            self.assertTrue(form.is_valid())
            self.assertEqual(
                [{'exposure_count': valid_form_data['exposure_count'],
                  'exposure_time': valid_form_data['exposure_time'],
                  'rotator_mode': 'VFLOAT', 'extra_params': {'rotator_angle': valid_form_data['rotator_angle']}
                  }], form._build_instrument_config()
            )


class TestLCOPhotometricSequenceForm(TestCase):

    def setUp(self):
        self.st = SiderealTargetFactory.create()
        self.valid_form_data = {
            'name': 'test', 'facility': 'LCO', 'target_id': self.st.id, 'ipp_value': 0.5, 'max_airmass': 3,
            'min_lunar_distance': 20, 'observation_mode': 'NORMAL', 'proposal': 'sampleproposal',
            'instrument_type': '0M4-SCICAM-SBIG', 'cadence_frequency': 24,
            'U_0': 30.0, 'U_1': 1, 'U_2': 1, 'B_0': 60.0, 'B_1': 2, 'B_2': 1,
        }
        self.instrument_choices = [(k, v['name']) for k, v in instrument_response.items() if 'SOAR' not in k]
        self.filter_choices = set([
            (f['code'], f['name']) for ins in instrument_response.values() for f in
            ins['optical_elements'].get('filters', [])
            if f['code'] in LCOPhotometricSequenceForm.valid_filters]
        )
        self.proposal_choices = generate_lco_proposal_choices()

    @patch('tom_observations.facilities.lco.LCOPhotometricSequenceForm._get_instruments')
    def test_instrument_choices(self, mock_get_instruments):
        """Test LCOPhotometricSequenceForm._instrument_choices."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()

        inst_choices = LCOPhotometricSequenceForm.instrument_choices()
        self.assertIn(('0M4-SCICAM-SBIG', '0.4 meter SBIG'), inst_choices)
        self.assertNotIn(('2M0-FLOYDS-SCICAM', '2.0 meter FLOYDS'), inst_choices)
        self.assertEqual(len(inst_choices), 1)

    @patch('tom_observations.facilities.lco.LCOPhotometricSequenceForm.proposal_choices')
    @patch('tom_observations.facilities.lco.LCOPhotometricSequenceForm.filter_choices')
    @patch('tom_observations.facilities.lco.LCOPhotometricSequenceForm.instrument_choices')
    @patch('tom_observations.facilities.lco.LCOPhotometricSequenceForm.validate_at_facility')
    def test_build_instrument_config(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        form = LCOPhotometricSequenceForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        inst_config = form._build_instrument_config()
        self.assertEqual(len(inst_config), 2)
        self.assertIn({'exposure_count': 1, 'exposure_time': 30.0, 'optical_elements': {'filter': 'U'}}, inst_config)
        self.assertIn({'exposure_count': 2, 'exposure_time': 60.0, 'optical_elements': {'filter': 'B'}}, inst_config)

    @patch('tom_observations.facilities.lco.LCOPhotometricSequenceForm.proposal_choices')
    @patch('tom_observations.facilities.lco.LCOPhotometricSequenceForm.filter_choices')
    @patch('tom_observations.facilities.lco.LCOPhotometricSequenceForm.instrument_choices')
    @patch('tom_observations.facilities.lco.LCOPhotometricSequenceForm.validate_at_facility')
    def test_clean(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        # Test that a valid form returns True, and that start and end are cleaned properly
        form = LCOPhotometricSequenceForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        self.assertAlmostEqual(datetime.strftime(datetime.now(), '%Y-%m-%dT%H:%M:%S'), form.cleaned_data['start'])
        self.assertAlmostEqual(
            datetime.strftime(
                datetime.now() + timedelta(hours=form.cleaned_data['cadence_frequency']), '%Y-%m-%dT%H:%M:%S'
            ),
            form.cleaned_data['end']
        )

        # Test that an invalid form returns False
        self.valid_form_data.pop('target_id')
        form = LCOPhotometricSequenceForm(self.valid_form_data)
        self.assertFalse(form.is_valid())


class TestLCOSpectroscopicSequenceForm(TestCase):
    def setUp(self):
        self.st = SiderealTargetFactory.create()
        self.valid_form_data = {
            'name': 'test', 'facility': 'LCO', 'target_id': self.st.id, 'exposure_count': 1, 'exposure_time': 30,
            'max_airmass': 3, 'min_lunar_distance': 20, 'site': 'any', 'ipp_value': 0.5, 'filter': 'slit_1.2as',
            'observation_mode': 'NORMAL', 'proposal': 'sampleproposal', 'acquisition_radius': 1,
            'guider_mode': 'on', 'guider_exposure_time': 30, 'instrument_type': '0M4-SCICAM-SBIG',
            'cadence_frequency': 24
        }
        self.instrument_choices = [(k, v['name']) for k, v in instrument_response.items() if 'SOAR' not in k]
        self.filter_choices = set([
            (f['code'], f['name']) for ins in instrument_response.values() for f in
            ins['optical_elements'].get('filters', []) + ins['optical_elements'].get('slits', [])
        ])
        self.proposal_choices = generate_lco_proposal_choices()

    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm._get_instruments')
    def test_instrument_choices(self, mock_get_instruments):
        """Test LCOSpectroscopicSequenceForm._instrument_choices."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()

        inst_choices = LCOSpectroscopicSequenceForm.instrument_choices()
        self.assertIn(('2M0-FLOYDS-SCICAM', '2.0 meter FLOYDS'), inst_choices)
        self.assertNotIn(('0M4-SCICAM-SBIG', '0.4 meter SBIG'), inst_choices)
        self.assertEqual(len(inst_choices), 1)

    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm._get_instruments')
    def test_filter_choices(self, mock_get_instruments):
        """Test LCOSpectroscopicSequenceForm._instrument_choices."""
        mock_get_instruments.return_value = generate_lco_instrument_choices()

        filter_choices = LCOSpectroscopicSequenceForm.filter_choices()
        for expected in [('slit_6.0as', '6.0 arcsec slit'), ('slit_1.6as', '1.6 arcsec slit'),
                         ('slit_2.0as', '2.0 arcsec slit'), ('slit_1.2as', '1.2 arcsec slit')]:
            self.assertIn(expected, filter_choices)
        for not_expected in [('opaque', 'Opaque'), ('100um-Pinhole', '100um Pinhole')]:
            self.assertNotIn(not_expected, filter_choices)
        self.assertEqual(len(filter_choices), 4)

    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.proposal_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.filter_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.instrument_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.validate_at_facility')
    def test_build_instrument_config(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        form = LCOSpectroscopicSequenceForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        inst_config = form._build_instrument_config()
        self.assertEqual(len(inst_config), 1)
        self.assertIn({'exposure_count': 1, 'exposure_time': 30.0, 'optical_elements': {'slit': 'slit_1.2as'}},
                      inst_config)
        self.assertNotIn('filter', inst_config[0]['optical_elements'])

    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.proposal_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.filter_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.instrument_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.validate_at_facility')
    def test_build_acquisition_config(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        with self.subTest():
            form = LCOSpectroscopicSequenceForm(self.valid_form_data)
            self.assertTrue(form.is_valid())
            acquisition_config = form._build_acquisition_config()
            self.assertDictEqual({'mode': 'BRIGHTEST', 'extra_params': {'acquire_radius': 1}},
                                 acquisition_config)

        with self.subTest():
            self.valid_form_data.pop('acquisition_radius')
            form = LCOSpectroscopicSequenceForm(self.valid_form_data)
            self.assertTrue(form.is_valid())
            acquisition_config = form._build_acquisition_config()
            self.assertDictEqual({'mode': 'WCS'}, acquisition_config)

    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.proposal_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.filter_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.instrument_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.validate_at_facility')
    def test_build_guiding_config(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        test_params = [
            ({'guider_mode': 'on'}, {'mode': 'ON', 'optional': 'false'}),
            ({'guider_mode': 'off'}, {'mode': 'OFF', 'optional': 'false'}),
            ({'guider_mode': 'optional'}, {'mode': 'ON', 'optional': 'true'})
        ]
        for params in test_params:
            with self.subTest():
                self.valid_form_data.update(params[0])
                form = LCOSpectroscopicSequenceForm(self.valid_form_data)
                self.assertTrue(form.is_valid())
                guiding_config = form._build_guiding_config()
                self.assertDictEqual(params[1], guiding_config)

    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.proposal_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.filter_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.instrument_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.validate_at_facility')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm._get_instruments')
    def test_build_location(self, mock_get_instruments, mock_validate, mock_insts, mock_filters, mock_proposals):
        mock_get_instruments.return_value = generate_lco_instrument_choices()
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        test_params = [
            ({'site': 'ogg'}, {'site': 'ogg'}),
            ({'site': 'coj'}, {'site': 'coj'}),
            ({'site': 'any'}, {})
        ]
        for params in test_params:
            with self.subTest():
                self.valid_form_data.update(params[0])
                form = LCOSpectroscopicSequenceForm(self.valid_form_data)
                self.assertTrue(form.is_valid())
                location = form._build_location()
                self.assertDictContainsSubset(params[1], location)

    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.proposal_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.filter_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.instrument_choices')
    @patch('tom_observations.facilities.lco.LCOSpectroscopicSequenceForm.validate_at_facility')
    def test_clean(self, mock_validate, mock_insts, mock_filters, mock_proposals):
        mock_validate.return_value = []
        mock_insts.return_value = self.instrument_choices
        mock_filters.return_value = self.filter_choices
        mock_proposals.return_value = self.proposal_choices

        # Test that a valid form returns True, and that start and end are cleaned properly
        form = LCOSpectroscopicSequenceForm(self.valid_form_data)
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['instrument_type'], '2M0-FLOYDS-SCICAM')
        self.assertAlmostEqual(datetime.strftime(datetime.now(), '%Y-%m-%dT%H:%M:%S'), form.cleaned_data['start'])
        self.assertAlmostEqual(
            datetime.strftime(
                datetime.now() + timedelta(hours=form.cleaned_data['cadence_frequency']), '%Y-%m-%dT%H:%M:%S'
            ),
            form.cleaned_data['end']
        )

        # Test that an invalid form returns False
        self.valid_form_data.pop('target_id')
        form = LCOSpectroscopicSequenceForm(self.valid_form_data)
        self.assertFalse(form.is_valid())


class TestLCOObservationTemplateForm(TestCase):
    pass


class TestLCOFacility(TestCase):
    def setUp(self):
        self.lco = LCOFacility()

    @patch('tom_observations.facilities.lco.make_request')
    def test_get_requestgroup_id(self, mock_make_request):
        mock_response = Response()
        mock_response._content = str.encode(json.dumps({
            'count': 1,
            'results': [{
                'id': 1073496
            }]
        }))
        mock_response.status_code = 200
        mock_make_request.return_value = mock_response

        with self.subTest('Test that a correct response results in a valid id.'):
            requestgroup_id = self.lco._get_requestgroup_id(1234567)
            self.assertEqual(requestgroup_id, 1073496)

        with self.subTest('Test that an empty response results in no id.'):
            mock_response._content = str.encode(json.dumps({
                'count': 0,
                'results': []
            }))
            requestgroup_id = self.lco._get_requestgroup_id(1234567)
            self.assertIsNone(requestgroup_id)

        with self.subTest('Test that multiple results returns no id.'):
            mock_response._content = str.encode(json.dumps({
                'count': 2,
                'results': [{'id': 1073496}, {'id': 1073497}]
            }))
            requestgroup_id = self.lco._get_requestgroup_id(1234567)
            self.assertIsNone(requestgroup_id)



import logging
from importlib import import_module

from django.conf import settings

logger = logging.getLogger(__name__)


def import_method(name):
    mod_name, method = name.rsplit('.', 1)
    try:
        mod = import_module(mod_name)
        m = getattr(mod, method)
    except (ImportError, AttributeError):
        raise ImportError('Could not import {}. Did you provide the correct path?'.format(name))
    return m


def run_hook(name, *args, **kwargs):
    hook = settings.HOOKS.get(name)
    if hook:
        method = import_method(hook)
        return method(*args, **kwargs)


def target_post_save(target, created):
    """This hook runs following update of a target."""
    logger.info('Target post save hook: %s created: %s', target, created)


def observation_change_state(observation, previous_state):
    """This hook runs upon the status of an observation changing."""
    logger.info('Observation change state hook: %s from %s to %s', observation, previous_state, observation.status)



from django import forms
from django.contrib.auth.models import Group
from django.conf import settings

from tom_dataproducts.models import DataProductGroup, DataProduct
from tom_observations.models import ObservationRecord
from tom_targets.models import Target


class AddProductToGroupForm(forms.Form):
    products = forms.ModelMultipleChoiceField(
        DataProduct.objects.all(),
        widget=forms.CheckboxSelectMultiple
    )
    group = forms.ModelChoiceField(DataProductGroup.objects.all())


class DataProductUploadForm(forms.Form):
    observation_record = forms.ModelChoiceField(
        ObservationRecord.objects.all(),
        widget=forms.HiddenInput(),
        required=False
    )
    target = forms.ModelChoiceField(
        Target.objects.all(),
        widget=forms.HiddenInput(),
        required=False
    )
    files = forms.FileField(
        widget=forms.ClearableFileInput(
            attrs={'multiple': True}
        )
    )
    data_product_type = forms.ChoiceField(
        choices=[v for k, v in settings.DATA_PRODUCT_TYPES.items()],
        widget=forms.RadioSelect(),
        required=True
    )
    referrer = forms.CharField(
        widget=forms.HiddenInput()
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not settings.TARGET_PERMISSIONS_ONLY:
            self.fields['groups'] = forms.ModelMultipleChoiceField(Group.objects.none(),
                                                                   required=False,
                                                                   widget=forms.CheckboxSelectMultiple)



from datetime import datetime, timezone
import json
from requests import Response
from unittest.mock import patch

from django.test import tag, TestCase
from faker import Faker

from tom_alerts.brokers.alerce import ALeRCEBroker, ALeRCEQueryForm
from tom_targets.models import Target


alerce_classifiers_response = [
    {
        'classifier_name': 'lc_classifier',
        'classifier_version': 'hierarchical_random_forest_1.0.0',
        'classes': ['SNIa', 'QSO', 'LPV']
    },
    {
        'classifier_name': 'lc_classifier_top',
        'classifier_version': 'hierarchical_random_forest_1.0.0',
        'classes': ['Transient', 'Stochastic', 'Periodic']
    },
    {
        'classifier_name': 'lc_classifier_transient',
        'classifier_version': 'hierarchical_random_forest_1.0.0',
        'classes': ['SNIa']
    },
    {
        'classifier_name': 'lc_classifier_stochastic',
        'classifier_version': 'hierarchical_random_forest_1.0.0',
        'classes': ['QSO']
    },
    {
        'classifier_name': 'lc_classifier_periodic',
        'classifier_version': 'hierarchical_random_forest_1.0.0',
        'classes': ['LPV']
    },
    {
        'classifier_name': 'stamp_classifier',
        'classifier_version': 'stamp_classifier_1.0.4',
        'classes': ['SN']
    },
]


def create_alerce_alert(firstmjd=None, lastmjd=None, class_name=None, classifier=None, probability=None):
    fake = Faker()

    return {'oid': fake.pystr_format(string_format='ZTF##???????', letters='abcdefghijklmnopqrstuvwxyz'),
            'meanra': fake.pyfloat(min_value=0, max_value=360),
            'meandec': fake.pyfloat(min_value=0, max_value=360),
            'firstmjd': firstmjd if firstmjd else fake.pyfloat(min_value=56000, max_value=59000, right_digits=1),
            'lastmjd': lastmjd if lastmjd else fake.pyfloat(min_value=56000, max_value=59000, right_digits=1),
            'class': class_name if class_name else fake.pystr_format(string_format='????????',
                                                                     letters='ABCDEFGHIJKLMNOPQRSTUVWXYZ'),
            'classifier': classifier if classifier else fake.pystr_format(string_format='???',
                                                                          letters='ABCDEFGHIJKLMNOPQRSTUVWXYZ'),
            'probability': probability if probability else fake.pyfloat(min_value=0, max_value=1)}


def create_alerce_query_response(num_alerts, page=1):
    alerts = [create_alerce_alert() for i in range(0, num_alerts)]

    return {
        'total': None, 'page': page, 'next': None, 'has_next': False, 'prev': None, 'has_prev': False,
        'items': alerts
    }


class TestALeRCEBrokerForm(TestCase):
    def setUp(self):
        self.base_form_data = {
            'query_name': 'Test Query',
            'broker': 'ALeRCE'
        }

    @patch('tom_alerts.brokers.alerce.cache')
    def test_cone_search_validation(self, mock_cache):
        """Test cross-field validation for cone search filters."""

        # Test that validation fails if not all fields are present
        parameters_list = [
            {'ra': 10, 'dec': 10}, {'dec': 10, 'radius': 10}, {'ra': 10, 'radius': 10}
        ]
        for parameters in parameters_list:
            with self.subTest():
                parameters.update(self.base_form_data)
                form = ALeRCEQueryForm(parameters)
                self.assertFalse(form.is_valid())
                self.assertIn('All of RA, Dec, and Search Radius must be included to execute a cone search.',
                              form.non_field_errors())

        # Test that validation passes when all three fields are present
        self.base_form_data.update({'ra': 10, 'dec': 10, 'radius': 10})
        form = ALeRCEQueryForm(self.base_form_data)
        self.assertTrue(form.is_valid())

    @patch('tom_alerts.brokers.alerce.cache')
    def test_time_filters_validation(self, mock_cache):
        """Test validation for time filters."""

        # Test that form validation succeeds when relative time fields make sense.
        parameters_list = [
            {'firstmjd__gt': 57000, 'firstmjd__lt': 58000, 'lastmjd__gt': 58000, 'lastmjd__lt': 59000},
        ]
        for parameters in parameters_list:
            with self.subTest():
                parameters.update(self.base_form_data)
                form = ALeRCEQueryForm(parameters)
                self.assertTrue(form.is_valid())

    @patch('tom_alerts.brokers.alerce.cache.get')
    def test_classifier_filters_validation(self, mock_cache_get):
        mock_cache_get.return_value = alerce_classifiers_response

        parameters_list = [
            {'lc_classifier': 'SNIa', 'p_lc_classifier': 0.5},
            {'stamp_classifier': 'SN', 'p_stamp_classifier': 0.5}
        ]
        for parameters in parameters_list:
            with self.subTest():
                parameters.update(self.base_form_data)
                form = ALeRCEQueryForm(parameters)
                form.is_valid()
                self.assertTrue(form.is_valid())

        invalid_parameters_list = [
            {'lc_classifier': 'SNIa', 'p_stamp_classifier': 0.5},
            {'stamp_classifier': 'SN', 'p_lc_classifier': 0.5}
        ]
        for parameters in invalid_parameters_list:
            with self.subTest():
                parameters.update(self.base_form_data)
                form = ALeRCEQueryForm(parameters)
                self.assertFalse(form.is_valid())
                self.assertIn('Only one of either light curve or stamp classification may be used as a filter.',
                              form.errors['__all__'])

    @patch('tom_alerts.brokers.alerce.cache.get')
    @patch('tom_alerts.brokers.alerce.requests.get')
    def test_get_classifiers(self, mock_requests_get, mock_cache_get):
        mock_response = Response()
        mock_response._content = str.encode(json.dumps(alerce_classifiers_response))
        mock_response.status_code = 200
        mock_requests_get.return_value = mock_response

        with self.subTest('Test that cached response avoids request.'):
            mock_cache_get.return_value = alerce_classifiers_response
            classifiers = ALeRCEQueryForm._get_classifiers()
            mock_requests_get.assert_not_called()
            for classifier_group in classifiers:
                self.assertTrue(
                    all(k in classifier_group.keys() for k in ['classifier_name', 'classifier_version', 'classes'])
                )

        with self.subTest('Test that no cached response results in HTTP request.'):
            mock_cache_get.return_value = None
            classifiers = ALeRCEQueryForm._get_classifiers()
            mock_requests_get.assert_called_once()

    @patch('tom_alerts.brokers.alerce.cache.get')
    def test_get_light_curve_classifier_choices(self, mock_cache_get):
        mock_cache_get.return_value = alerce_classifiers_response
        lc_classifiers = ALeRCEQueryForm._get_light_curve_classifier_choices()
        expected_classifiers = [
            (None, ''),
            ('SNIa', 'SNIa - transient'),
            ('QSO', 'QSO - stochastic'),
            ('LPV', 'LPV - periodic')
        ]
        for classifier in expected_classifiers:
            self.assertIn(classifier, lc_classifiers)

    @patch('tom_alerts.brokers.alerce.cache.get')
    def test_get_stamp_classifier_choices(self, mock_cache_get):
        mock_cache_get.return_value = alerce_classifiers_response
        stamp_classifiers = ALeRCEQueryForm._get_stamp_classifier_choices()
        expected_classifiers = {
            (None, ''),
            ('SN', 'SN')
        }
        for classifier in expected_classifiers:
            self.assertIn(classifier, stamp_classifiers)


class TestALeRCEBrokerClass(TestCase):
    def setUp(self):
        self.base_form_data = {
            'query_name': 'Test ALeRCE',
            'broker': 'ALeRCE',
        }
        self.broker = ALeRCEBroker()

    def test_clean_coordinate_parameters(self):
        """Test that _clean_date_parameters results in the correct dict structure."""
        parameters_list = [
            ({'ra': 10, 'dec': 10, 'radius': None}, []),
            ({'ra': 10, 'dec': 10, 'radius': 10}, [('ra', 10), ('dec', 10), ('radius', 10)])
        ]
        for parameters, expected_list in parameters_list:
            with self.subTest():
                cleaned_coordinate_parameters = self.broker._clean_coordinate_parameters(parameters)
                for expected in expected_list:
                    self.assertIn(expected, cleaned_coordinate_parameters)

    def test_clean_date_parameters(self):
        """Test that _clean_date_parameters results in the correct dict structure."""
        parameters_list = [
            ({'firstmjd__gt': 57000, 'firstmjd__lt': 58000, 'lastmjd__gt': 58000, 'lastmjd__lt': 59000},
             [('firstmjd', 57000), ('firstmjd', 58000), ('lastmjd', 58000), ('lastmjd', 59000)]),
            ({'firstmjd__gt': 57000, 'firstmjd__lt': 58000, 'lastmjd__gt': None, 'lastmjd__lt': None},
             [('firstmjd', 57000), ('firstmjd', 58000)]),
            ({'firstmjd__gt': None, 'firstmjd__lt': None, 'lastmjd__gt': None, 'lastmjd__lt': None}, [])
        ]
        for parameters, expected_list in parameters_list:
            with self.subTest():
                cleaned_date_parameters = self.broker._clean_date_parameters(parameters)
                for expected in expected_list:
                    self.assertIn(expected, cleaned_date_parameters)

    def test_clean_classifier_parameters(self):
        """Test that _clean_filter_parameters results in the correct dict structure."""

        # Test that classifiers are populated correctly
        parameters_list = [
            ({'stamp_classifier': None, 'p_stamp_classifier': None, 'lc_classifier': None, 'p_lc_classifier': None},
             []),
            ({'stamp_classifier': 'SN', 'p_stamp_classifier': None, 'lc_classifier': None, 'p_lc_classifier': None},
             [('classifier', 'stamp_classifier'), ('class', 'SN')]),
            ({'stamp_classifier': 'SN', 'p_stamp_classifier': 0.5, 'lc_classifier': None, 'p_lc_classifier': None},
             [('classifier', 'stamp_classifier'), ('class', 'SN'), ('probability', 0.5)]),
            ({'stamp_classifier': None, 'p_stamp_classifier': None, 'lc_classifier': 'SNIa', 'p_lc_classifier': None},
             [('classifier', 'lc_classifier'), ('class', 'SNIa')]),
            ({'stamp_classifier': None, 'p_stamp_classifier': None, 'lc_classifier': 'SNIa', 'p_lc_classifier': 0.5},
             [('classifier', 'lc_classifier'), ('class', 'SNIa'), ('probability', 0.5)]),
        ]
        for parameters, expected_list in parameters_list:
            with self.subTest():
                for expected in expected_list:
                    cleaned_classifier_parameters = self.broker._clean_classifier_parameters(parameters)
                    self.assertIn(expected, cleaned_classifier_parameters)

    @patch('tom_alerts.brokers.alerce.ALeRCEBroker._clean_classifier_parameters')
    @patch('tom_alerts.brokers.alerce.ALeRCEBroker._clean_date_parameters')
    @patch('tom_alerts.brokers.alerce.ALeRCEBroker._clean_coordinate_parameters')
    def test_clean_parameters(self, mock_coordinate, mock_date, mock_classifier):
        mock_coordinate.return_value = [('ra', 10), ('dec', 10), ('radius', 10)]
        mock_date.return_value = [('firstmjd', 57000), ('firstmjd', 58000), ('lastmjd', 58000), ('lastmjd', 59000)]
        mock_classifier.return_value = [('classifier', 'stamp_classifier'), ('class', 'SN'), ('probability', 0.5)]

        # Ensure that passed in values are used to populate the payload
        parameters = {'page': 2, 'oid': 'testoid', 'ndet': 10, 'ranking': 1, 'order_by': 'oid', 'order_mode': 'ASC'}
        payload = self.broker._clean_parameters(parameters)
        with self.subTest():
            self.assertIn(('page', 2), payload)
            self.assertIn(('page_size', 20), payload)
            for k, v in parameters.items():
                self.assertIn((k, v), payload)
            for expected in mock_coordinate.return_value:
                self.assertIn(expected, payload)
            for expected in mock_date.return_value:
                self.assertIn(expected, payload)
            for expected in mock_classifier.return_value:
                self.assertIn(expected, payload)

        # Ensure that missing values result in default values being used to populate the payload
        payload = self.broker._clean_parameters({})
        with self.subTest():
            self.assertIn(('page', 1), payload)

    @patch('tom_alerts.brokers.alerce.requests.get')
    @patch('tom_alerts.brokers.alerce.ALeRCEBroker._clean_parameters')
    def test_fetch_alerts(self, mock_clean_parameters, mock_requests_get):
        """Test fetch_alerts broker method."""
        first_mock_response_content = create_alerce_query_response(20, page=1)
        first_mock_response = Response()
        first_mock_response._content = str.encode(json.dumps(first_mock_response_content))
        first_mock_response.status_code = 200

        second_mock_response = Response()
        second_mock_response._content = str.encode(json.dumps(create_alerce_query_response(5, page=2)))
        second_mock_response.status_code = 200

        third_mock_response = Response()
        third_mock_response._content = str.encode(json.dumps(create_alerce_query_response(0, page=3)))
        third_mock_response.status_code = 200

        with self.subTest():
            mock_requests_get.side_effect = [first_mock_response, second_mock_response, third_mock_response]
            response = self.broker.fetch_alerts({'max_pages': 5})
            alerts = []
            for alert in response:
                alerts.append(alert)
            self.assertEqual(25, len(alerts))
            self.assertEqual(alerts[0], first_mock_response_content['items'][0])

        with self.subTest():
            mock_requests_get.side_effect = [first_mock_response]
            response = self.broker.fetch_alerts({'max_pages': 1})
            alerts = []
            for alert in response:
                alerts.append(alert)
            self.assertEqual(20, len(alerts))

    @patch('tom_alerts.brokers.alerce.requests.get')
    def test_fetch_alert(self, mock_requests_post):
        """Test fetch_alert broker method."""
        alert = create_alerce_alert(1)
        mock_response = Response()
        mock_response_content = alert
        mock_response._content = str.encode(json.dumps(mock_response_content))
        mock_response.status_code = 200
        mock_requests_post.return_value = mock_response

        alert_response = self.broker.fetch_alert(alert['oid'])
        self.assertDictEqual(alert_response, alert)

    def test_to_target(self):
        """Test to_target broker method."""
        mock_alert = create_alerce_alert()
        self.broker.to_target(mock_alert)
        t = Target.objects.first()

        self.assertEqual(mock_alert['oid'], t.name)
        self.assertEqual(mock_alert['meanra'], t.ra)
        self.assertEqual(mock_alert['meandec'], t.dec)

    def test_to_generic_alert(self):
        """Test to_generic_alert broker method."""

        # Test that timestamp is populated correctly.
        mock_alert = create_alerce_alert()
        mock_alert['lastmjd'] = None
        self.assertEqual('', self.broker.to_generic_alert(mock_alert).timestamp)

        mock_alert = create_alerce_alert(lastmjd=59155)
        self.assertEqual(datetime(2020, 11, 2, tzinfo=timezone.utc),
                         self.broker.to_generic_alert(mock_alert).timestamp)

        # Test that the url is created properly.
        mock_alert = create_alerce_alert()
        self.assertEqual(f'https://alerce.online/object/{mock_alert["oid"]}',
                         self.broker.to_generic_alert(mock_alert).url)

        # Test that the classification is selected correctly
        mock_alert = create_alerce_alert()
        self.assertEqual(mock_alert['probability'], self.broker.to_generic_alert(mock_alert).score)

        mock_alert = create_alerce_alert()
        mock_alert['probability'] = None
        self.assertEqual(None, self.broker.to_generic_alert(mock_alert).score)


@tag('canary')
class TestALeRCEModuleCanary(TestCase):
    def setUp(self):
        self.broker = ALeRCEBroker()
        self.base_form_parameters = {
            'query_name': 'Test ALeRCE',
            'broker': 'ALeRCE',
        }

    @patch('tom_alerts.brokers.alerce.cache.get')
    def test_get_classifiers(self, mock_cache_get):
        mock_cache_get.return_value = None  # Ensure cache is not used

        expected_classifiers = ['lc_classifier', 'stamp_classifier']
        classifiers = ALeRCEQueryForm._get_classifiers()
        for expected in expected_classifiers:
            for classifier in classifiers:
                if expected == classifier['classifier_name']:
                    break
            else:
                self.fail(f'Did not find {expected} in classifiers.')

    # These tests fail when running from Github Actions
    # def test_fetch_alerts(self):
    #     form = ALeRCEQueryForm(self.base_form_parameters)
    #     form.is_valid()
    #     query = form.save()

    #     alerts = [alert for alert in self.broker.fetch_alerts(query.parameters)]

    #     self.assertGreaterEqual(len(alerts), 1)
    #     for k in ['oid', 'firstmjd', 'lastmjd', 'class', 'classifier', 'probability', 'meanra', 'meandec']:
    #         self.assertIn(k, alerts[0])

    # def test_fetch_alerts_cone_search(self):
    #     parameters = {'ra': 174.5, 'dec': 5.5, 'radius': 240}
    #     parameters.update(self.base_form_parameters)
    #     form = ALeRCEQueryForm(parameters)
    #     form.is_valid()
    #     query = form.save()

    #     alerts = [alert for alert in self.broker.fetch_alerts(query.parameters)]

    #     self.assertGreaterEqual(len(alerts), 1)
    #     for alert in alerts:
    #         self.assertAlmostEqual(alert['meanra'], 174.5, 0)  # Test that RA is near enough to 174 to be valid
    #         self.assertAlmostEqual(alert['meandec'], 5.5, 0)  # Test that Declination is near enough to 5 to be valid

    # def test_fetch_alerts_classification_search(self):
    #     parameters_list = [
    #         ({'lc_classifier': 'SNIa', 'p_lc_classifier': 0.5},
    #          {'class': 'SNIa', 'classifier': 'lc_classifier', 'probability': 0.5}),
    #         ({'stamp_classifier': 'SN', 'p_stamp_classifier': 0.5},
    #          {'class': 'SN', 'classifier': 'stamp_classifier', 'probability': 0.5})
    #     ]

    #     for parameters, expected in parameters_list:
    #         with self.subTest():
    #             parameters.update(self.base_form_parameters)
    #             form = ALeRCEQueryForm(parameters)
    #             form.is_valid()
    #             query = form.save()

    #             alerts = [alert for alert in self.broker.fetch_alerts(query.parameters)]

    #             self.assertGreaterEqual(len(alerts), 1)
    #             for alert in alerts:
    #                 self.assertEqual(alert['class'], expected['class'])
    #                 self.assertEqual(alert['classifier'], expected['classifier'])
    #                 self.assertGreaterEqual(alert['probability'], expected['probability'])

    # def test_fetch_alerts_time_filters(self):
    #     parameters = {'firstmjd__gt': 59000, 'firstmjd__lt': 59100, 'lastmjd__gt': 59300, 'lastmjd__lt': 59400}
    #     parameters.update(self.base_form_parameters)
    #     form = ALeRCEQueryForm(parameters)
    #     form.is_valid()
    #     query = form.save()

    #     alerts = [alert for alert in self.broker.fetch_alerts(query.parameters)]

    #     self.assertGreaterEqual(len(alerts), 1)
    #     for alert in alerts:
    #         self.assertGreaterEqual(alert['firstmjd'], 59000)
    #         self.assertLessEqual(alert['firstmjd'], 59100)
    #         self.assertGreaterEqual(alert['lastmjd'], 59300)
    #         self.assertLessEqual(alert['lastmjd'], 59400)

    # def test_fetch_alerts_other_filters(self):
    #     parameters = {'ndet': 10}
    #     parameters.update(self.base_form_parameters)
    #     form = ALeRCEQueryForm(parameters)
    #     form.is_valid()
    #     query = form.save()

    #     alerts = [alert for alert in self.broker.fetch_alerts(query.parameters)]

    #     self.assertGreaterEqual(len(alerts), 1)
    #     for alert in alerts:
    #         self.assertGreaterEqual(alert['ndet'], 10)

    # def test_ordering(self):
    #     parameters = {'lc_classifier': 'SNIa'}
    #     parameters.update(self.base_form_parameters)
    #     sorting_parameters = ['oid', 'probability', 'ndet', 'firstmjd', 'lastmjd', 'meanra', 'meandec']
    #     sort_ordering_parameters = ['ASC', 'DESC']

    #     for sorting_parameter in sorting_parameters:
    #         for sort_order in sort_ordering_parameters:
    #             if sorting_parameters != 'ndet' and sort_order != 'ASC':  # This specific combination results in a 500
    #                 with self.subTest():
    #                     parameters.update(self.base_form_parameters)
    #                     parameters.update({'order_by': sorting_parameter, 'order_mode': sort_order})
    #                     form = ALeRCEQueryForm(parameters)
    #                     form.is_valid()
    #                     query = form.save()

    #                     alerts = [alert for alert in self.broker.fetch_alerts(query.parameters)]
    #                     self.assertGreaterEqual(len(alerts), 2)

    #                     last_alert = None
    #                     for alert in alerts:
    #                         if last_alert:
    #                             if sort_order == 'ASC':
    #                                 self.assertGreaterEqual(alert[sorting_parameter], last_alert[sorting_parameter])
    #                             elif sort_order == 'DESC':
    #                                 self.assertLessEqual(alert[sorting_parameter], last_alert[sorting_parameter])

    def test_fetch_alert(self):
        """
        Test fetching a single alert from ALeRCE. The two response values tested are the only ones that are not derived,
        and are therefore consistent.
        """
        alert = self.broker.fetch_alert('ZTF20acnsdjd')

        self.assertDictContainsSubset({
            'oid': 'ZTF20acnsdjd',
            'firstmjd': 59149.1119328998,
        }, alert)



from requests import Response

from django.utils import timezone
from django.test import TestCase, override_settings
from django.forms import ValidationError
from unittest import mock

from tom_alerts.alerts import get_service_class
from tom_alerts.brokers.gaia import GaiaQueryForm
from tom_alerts.brokers.gaia import GaiaBroker
from tom_targets.models import Target
from tom_dataproducts.models import ReducedDatum


@override_settings(TOM_ALERT_CLASSES=['tom_alerts.brokers.gaia.GaiaBroker'])
class TestGaiaQueryForm(TestCase):
    def setUp(self):
        self.base_form_params = {'query_name': 'Test Query', 'broker': 'Gaia'}

    def test_with_required_params(self):
        self.base_form_params['target_name'] = 'Test Target'
        form = GaiaQueryForm(self.base_form_params)
        self.assertTrue(form.is_valid())

    def test_no_query_name(self):
        self.base_form_params['target_name'] = 'Test Target'
        self.base_form_params.pop('query_name')
        form = GaiaQueryForm(self.base_form_params)
        self.assertFalse(form.is_valid())
        self.assertIn('This field is required.', form.errors.get('query_name'))

    def test_no_target_name_or_cone(self):
        form = GaiaQueryForm(self.base_form_params)
        self.assertFalse(form.is_valid())
        with self.assertRaises(ValidationError):
            form.clean()
        self.assertIn('Please enter either a target name or cone search parameters.', form.errors.get('__all__'))

    def test_both_target_name_and_cone(self):
        self.base_form_params['target_name'] = 'Test Target'
        self.base_form_params['cone'] = '10,20,3'
        form = GaiaQueryForm(self.base_form_params)
        self.assertFalse(form.is_valid())
        with self.assertRaises(ValidationError):
            form.clean()
        self.assertIn('Please only enter one of target name or cone search parameters.', form.errors.get('__all__'))

    def test_cone(self):
        self.base_form_params['cone'] = '10,20,3'
        form = GaiaQueryForm(self.base_form_params)
        self.assertTrue(form.is_valid())

    def test_cone_invalid_format(self):
        self.base_form_params['cone'] = '10'
        form = GaiaQueryForm(self.base_form_params)
        self.assertFalse(form.is_valid())
        with self.assertRaises(ValidationError):
            form.clean()
        self.assertIn('Cone search parameters must be in the format \'RA,Dec,Radius\'.', form.errors.get('cone'))


@override_settings(TOM_ALERT_CLASSES=['tom_alerts.brokers.gaia.GaiaBroker'])
class TestGaiaBroker(TestCase):
    def setUp(self):
        self.test_html = """
            <html>
                <script charset="utf-8" type="text/javascript">
                    var alerts = [{"name": "Gaia20cpu", "tnsid": "AT2020lto", "obstime": "2020-06-04 17:25:08",
                        "ra": "291.61247", "dec": "13.36801", "alertMag": "20.54", "historicMag": "17.79",
                        "historicStdDev": "0.24", "classification": "CV", "published": "2020-06-06 12:26:16",
                        "comment": "test comment", "per_alert": {"link": "/alerts/alert/Gaia20cpu/", "name":
                        "Gaia20cpu"}, "rvs": false},
                        {"name": "Gaia20bph", "tnsid": "AT2020ftt", "obstime": "2020-04-01 12:52:23",
                        "ra": "34.02266", "dec": "68.65102", "alertMag": "16.21", "historicMag": "18.39",
                        "historicStdDev": "1.22", "classification": "unknown", "published": "2020-04-03 09:47:04",
                        "comment": "candidate CV; several previous outbursts in lightcurve",
                        "per_alert": {"link": "/alerts/alert/Gaia20bph/", "name": "Gaia20bph"}, "rvs": false}];

                    $(document).ready(function() {
                        var index_table = $('#alertsindex').dataTable( {
                            "data": alerts,
                        });
                    });
                </script>
            </html>"""
        self.test_html = self.test_html.replace('\n', '')
        self.alert_list = [
                            {
                                "name": "Gaia20cpu", "tnsid": "AT2020lto", "obstime": "2020-06-04 17:25:08",
                                "ra": "291.61247", "dec": "13.36801", "alertMag": "20.54", "historicMag": "17.79",
                                "historicStdDev": "0.24", "classification": "CV", "published": "2020-06-06 12:26:16",
                                "comment": "test comment", "per_alert": {
                                    "link": "/alerts/alert/Gaia20cpu/", "name": "Gaia20cpu"
                                }, "rvs": 'false'},
                            {
                                "name": "Gaia20bph", "tnsid": "AT2020ftt", "obstime": "2020-04-01 12:52:23",
                                "ra": "34.02266", "dec": "68.65102", "alertMag": "16.21", "historicMag": "18.39",
                                "historicStdDev": "1.22", "classification": "unknown",
                                "published": "2020-04-03 09:47:04",
                                "comment": "candidate CV; several previous outbursts in lightcurve", "per_alert": {
                                    "link": "/alerts/alert/Gaia20bph/", "name": "Gaia20bph"},
                                "rvs": 'false'}
                        ]
        self.test_target = Target.objects.create(name=self.alert_list[0]['name'])
        ReducedDatum.objects.create(
            source_name='Gaia',
            source_location=111111,
            target=self.test_target,
            data_type='photometry',
            timestamp=timezone.now(),
            value=12345.6789
        )

    @mock.patch('tom_alerts.brokers.gaia.requests.get')
    def test_fetch_alerts(self, mock_requests_get):
        mock_response = Response()
        mock_response._content = self.test_html
        mock_response.status_code = 200
        mock_requests_get.return_value = mock_response

        search_params = {'target_name': 'Gaia20bph', 'cone': None, }
        alerts = GaiaBroker().fetch_alerts(search_params)
        self.assertEqual(1, sum(1 for _ in alerts))

        search_params = {'target_name': None, 'cone': '291.61247, 13.36801, 0.002'}
        alerts = GaiaBroker().fetch_alerts(search_params)
        self.assertEqual(1, sum(1 for _ in alerts))

    def test_to_generic_alert(self):
        alert = GaiaBroker().to_generic_alert(self.alert_list[0])
        self.assertEqual(alert.name, self.alert_list[0]['name'])

    @mock.patch('tom_alerts.brokers.gaia.requests.get')
    def test_process_reduced_data_with_alert(self, mock_requests_get):

        mock_photometry_response = Response()
        mock_photometry_response._content = str.encode('''Gaia20bph\n#Date,JD,averagemag.\n
                                    2014-08-01 00:05:24,2456870.504,19.48\n2014-08-01 06:05:38,2456870.754,19.48\n\n''')
        mock_photometry_response.status_code = 200
        mock_requests_get.return_value = mock_photometry_response

        GaiaBroker().process_reduced_data(self.test_target, alert=self.alert_list[0])

        reduced_data = ReducedDatum.objects.filter(target=self.test_target, source_name='Gaia')
        self.assertGreater(reduced_data.count(), 1)

    @mock.patch('tom_alerts.brokers.gaia.requests.get')
    @mock.patch('tom_alerts.brokers.gaia.GaiaBroker.fetch_alerts')
    def test_process_reduced_data_without_alert(self, mock_fetch_alerts, mock_requests_get):
        mock_fetch_alerts.return_value = iter([self.alert_list[1]])

        mock_photometry_response = Response()
        mock_photometry_response._content = str.encode('''Gaia20bph\n#Date,JD,averagemag.\n
                                    2014-08-01 00:05:24,2456870.504,19.48\n2014-08-01 06:05:38,2456870.754,19.48\n\n''')
        mock_photometry_response.status_code = 200
        mock_requests_get.return_value = mock_photometry_response

        GaiaBroker().process_reduced_data(self.test_target)

        reduced_data = ReducedDatum.objects.filter(target=self.test_target, source_name='Gaia')
        self.assertGreater(reduced_data.count(), 1)

    def test_get_broker_class(self):
        self.assertEqual(GaiaBroker, get_service_class('Gaia'))



# Generated by Django 2.2.2 on 2019-07-15 16:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tom_targets', '0009_auto_20190605_1656'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='targetlist',
            options={'ordering': ('-created', 'name')},
        ),
    ]
